bug_check1 = """
    node state {
        has cand_intents = [];
        # Collect possible intents from a given conversation state
        can collect_intents {
            for i in -[transition]->.edge {
                here.cand_intents += [i.intent_label];
            }
        }
    }
    edge transition {has intent_label;}
    walker test_walker {
        state {
            here::collect_intents;
        }
    }
    graph test_graph {
        has anchor state_node;
        spawn {
            state_node = spawn node::state;
            state_2 = spawn node::state;
            state_node +[transition(intent_label =
                        "THIS IS AN INTENT_LABEL")]+> state_2;
        }
    }

    walker init {
        spawn here ++> graph::test_graph;
        spawn --> walker::test_walker;
        report -->[0].cand_intents;
        report -->;
    }

    test "Test"
    with graph::test_graph by walker::test_walker {
        assert(here.cand_intents == ['THIS IS AN INTENT_LABEL']);
    }
    """


action_load_std_lib = """
    walker aload {
        report std.actload_local('jaseci/tests/infer.py');
    }
    """


globals = """
    global a = "testing";

    walker init {
        report global.a;
        global.a = 56;
        report global.a;
    }
    """

net_root_std_lib = """
    walker init {
        root {
            report [here.info['jid'], net.root().info['jid']];
            spawn here ++> node::generic;
            take -->;
        }
        generic {
            report [here.info['jid'], net.root().info['jid']];
        }
    }
    """


or_stmt = """
    walker init {
        x = 3.4;
        y = "Hello";
        if(x == 3.45 or y == "Hello"):
            report [x, y];
    }
    """

nd_equals_error_correct_line = """
    node plain{ has name="joe", noname;
    can stuff with entry {std.out(name + noname); }}

    walker init {
        root {
            spawn here ++> node::plain;
            take -->;
        }
        plain {
            if(here.details['name'] == 'joe'):
                report here.info;
        }
    }
    """

strange_ability_bug = """
    node plain {
        can show with entry {
            report "Showing";
        }
    }

    walker init {
        root {
            spawn here ++> node::plain;
        }
    }

    walker travel {
        take -->;
    }
    """

node_inheritance = """
    node plain {
        has a=5, b=7, c=7, d=8;
        can x with entry {
            report "plain.x";
        }
        can y {
            report "plain.y";
        }
    }

    node plain2 {
        has c=70, d=80;
        can x with entry {
            report "plain2.x";
        }
        can y {
            report "plain2.y";
        }
    }

    node super:plain:plain2 {
        has a=55, c=7;
        can x with entry {
            ::plain:x;
            report here.context;
            report "super.x";
        }
        can y {
            ::plain2:y;
            report "super.y";
        }
    }

    walker init {
        root {
            a=spawn here ++> node::super;
        }
        take -->;
        super {
            here::y;
        }
    }
    """

inherited_ref = """
    node plain {
        has a=5, b=7, c=7, d=8;
        can x with entry {
            report "plain.x";
        }
        can y {
            report "plain.y";
        }
        can z with entry {
            report "Z triggered";
        }
    }

    node plain2:plain {
        has c=70, d=80;
        can x with entry {
            report "plain2.x";
        }
        can y {
            report "plain2.y";
        }
        can z with entry {
            report "New Z triggered";
        }
    }

    node super:plain2 {
        has a=55, c=7;
        can x with entry {
            ::plain:x;
            report here.context;
            report "super.x";
            ::z;
        }
        can y {
            ::plain2:y;
            report "super.y";
        }
    }

    walker init {
        root {
            spawn here ++> node::super;
            spawn here ++> node::plain;
            spawn here ++> node::plain2;
        }
        take --> node::plain;
        plain {
            report here;
        }
    }
    """

node_inheritance_chain_check = """
    node plain {
        has a=5, b=7, c=7, d=8;
        can x with entry {
            report "plain.x";
        }
        can y {
            report "plain.y";
        }
    }

    node plain2 {
        has c=70, d=80;
        can x with entry {
            report "plain2.x";
        }
        can y {
            report "plain2.y";
        }
    }

    node super:plain {
        has a=55, c=7;
        can x with entry {
            ::plain:x;
            report here.context;
            report "super.x";
        }
        can y {
            ::plain2:y;
            report "super.y";
        }
    }

    walker init {
        root {
            a=spawn here ++> node::super;
        }
        take -->;
        super {
            here::y;
        }
    }
    """

global_reregistering = """
    node plain;

    global a = '556';

    walker init {
        root {
            spawn here ++> node::plain;
            spawn here ++> node::plain;
            spawn here ++> node::plain;
        }
        report global.a;
        take -->;
    }
    """

vector_cos_sim_check = """
    node plain;

    walker init {
        a=[1,2,3];
        b=[4,5,6];
        report vector.cosine_sim(a,b);
    }
    """

multi_breaks = """
    node plain {
        has anchor val=0;
        can breakdance {
            for i=0 to i<10 by i+=1 {
                for j=0 to j<20 by j+=1:
                    if(j==12) {
                        val+=j;
                        break;
                    }
                report "here";
                if(i==6){
                    val+=i;
                    break;
                }
            }
            break;
            val+=100;
        }
    }

    walker init {
        nd=spawn here ++> node::plain;
        nd::breakdance;
        nd::breakdance;
        report nd.val;
    }
    """

reffy_deref_check = """
    node plain{has expected_answer;}

    walker init {
        nd = spawn here ++> node::plain;
        spawn here ++> node::plain;

        report *&-->[0] == *&-->[1];

        actual_answer = -->[1];
        nd.expected_answer = &-->[1];
        report *&actual_answer == *nd.expected_answer;
    }
    """

vanishing_can_check = """
    node plain {
        has name;
        can infer.year_from_date;
    }

    walker init {
        root {
            take --> node::plain else {
                nd=spawn here ++> node::plain;
                report nd.info['jid'];
                disengage;
            }
        }
        plain {
            report infer.year_from_date("2022-05-05");
        }
    }
    """

jasecilib_alias_list = """
    walker init {
        report jaseci.alias_list();
    }
    """

jasecilib_params = """
    walker init {
        report jaseci.graph_list(true);
    }
    """

jasecilib_create_user = """
    walker init {
        report jaseci.master_create("daman@gmail.com");
    }
    """

root_is_node_type = """
    walker init {
        report here.type;
    }
    """

walker_with_exit_after_node = """
    node echeck {
        has a=3;
        can dostuff with exit {
            report a;
        }
    }

    walker init {
        has a=0;
        with entry {
            a+=1;
        }
        root {
            spawn here ++> node::echeck;
            spawn here ++> node::echeck;
            spawn here ++> node::echeck;
            spawn here ++> node::echeck;
        }
        take -->;
        report a;
        with exit {
            report 43;
        }
    }
    """


depth_first_take = """
    node a {
        has num;
    }

    walker init {
        root {
            n1=spawn node::a(num=1);
            n2=spawn node::a(num=2);
            n3=spawn node::a(num=3);
            n4=spawn node::a(num=4);
            n5=spawn node::a(num=5);
            n6=spawn node::a(num=6);
            n7=spawn node::a(num=7);

            here ++> n1 ++> n2 ++> n3;
                            n2 ++> n4;
                     n1 ++> n5 ++> n6;
                            n5 ++> n7;
        }

        a: report here.num;
        take:dfs -->;
    }
    """

breadth_first_take = """
    node a {
        has num;
    }

    walker init {
        root {
            n1=spawn node::a(num=1);
            n2=spawn node::a(num=2);
            n3=spawn node::a(num=3);
            n4=spawn node::a(num=4);
            n5=spawn node::a(num=5);
            n6=spawn node::a(num=6);
            n7=spawn node::a(num=7);

            here ++> n1 ++> n2 ++> n3;
                            n2 ++> n4;
                     n1 ++> n5 ++> n6;
                            n5 ++> n7;
        }

        a: report here.num;
        take:bfs -->;
    }
    """


inheritance_override_here_check = """
    node a {
        has x=5, y=3, z=1;
        can sum {
            z = x+y+z;
            report z;
        }
    }

    node b:a {
        can sum {
            z = x-y-z;
            report x-y;
        }
    }

    node c:a {
        can sum {
            ::a:sum;
            z=z+1;
            report here.z;
        }
    }

    walker init {
        n1 = spawn here ++> node::a;
        n2 = spawn here ++> node::c;
        -->[0]::sum;
        -->[1]::sum;
    }
    """

dot_private_hidden = """
    node a {
        has x="56";
        has private j="5566669";
    }

    walker init {
        n1 = spawn here ++> node::a;
        n2 = spawn here ++> node::a;
    }
    """

check_destroy_node_has_var = """
    node a {
        has x;
    }

    walker create {
        n = spawn here ++> node::a;
        n.x = spawn node::a;
        report n.x.type;
    }

    walker remove {
        n=-->[0];
        destroy n.x;
        report n.x.type;
    }
"""

check_dict_for_in_loop = """
    walker for_loop_dict {
        with entry {
            testing = {
                "test1": 1,
                "test2": 2,
                "test3": 3
            };

            for key in testing {
                report key.str + " : " + testing[key].str;
            }

            for key, val in testing {
                report key.str + " : " + val.str;
            }

            testing = [5,6,7];

            for key in testing {
                report key;
            }

            for key, val in testing {
                report key.str + " : " + val.str;
            }
        }
    }

    walker var_as_key_for_dict {
        with entry {
            key = "key1";
            not_str_key = 1;
            testing = {
                key: key,
                "key2": 2,
                not_str_key: not_str_key
            };

            report testing;
        }
    }
"""

list_pairwise = """
walker init{
    with entry{
        _list = [1,2,3,4];
        p = _list.list::pairwise;
        report p;
    }
    }
"""

list_unique = """
walker init{
    with entry{
        _list = [1,2,3,4,2,3,4,5];
        p = _list.list::unique;
        report p;
    }
    }
"""

check_new_builtin = """
    walker init {
        with entry {
            a = {"test":"test"};

            // dict get with default if not existing
            b = a.dict::get("t", 1);

            // string join single param array
            c = " ".str::join([1,2,3,4]);

            // string join multiparams
            d = " ".str::join(1,2,3,4);
            report a;
            report b;
            report c;
            report d;
        }
    }
"""


continue_issue = """
    walker init {
        root {
            for i=0 to i<10 by i+=1 {
                if(i==9):
                    continue;
                if(i):
                    report i;
            }
            report "apple";
        }
    }
"""

async_syntax = """
async walker simple_async_second_layer {
    has anchor val1, val2 = 2;
    with entry {
        report val1;
        report val2;
    }
}

async walker simple_async {
    has anchor test_anchor = "test";
    can task.get_result;
    with entry {
        task1 = spawn here sync walker::simple_async_second_layer(val1 = 1);
        task2 = spawn here walker::simple_async_second_layer(val1 = 2);
        report task1;
        report task2;

        if task2["is_queued"]:
            report task.get_result(task2["result"], true);
    }
}

async walker simple_async_with_sync {
    has anchor test_anchor = "test";
    can task.get_result;
    with entry {
        task1 = spawn here sync walker::simple_async_second_layer(val1 = 1);
        task2 = spawn here walker::simple_async_second_layer(val1 = 2);
        report task1;
        report task2;

        report sync task2;
    }
}
"""

async_syntax_with_update = """
node a {
    has value = false;
}
walker init {
    root {
        spawn here ++> node::a;
    }
}

async walker update_value {
    root {
        take --> node::a;
    }
    a {
        here.value = true;
        report here.context;
    }
}

walker get_value {
    root {
        take --> node::a;
    }
    a {
        report here.context;
    }
}
"""

block_scope_check = """
    walker init {
        i=5;
        for i=0 to i<10 by i+=1 {
            report i;
        }
        report i;
    }
"""

ignore_check = """
    walker init {
        has nds=[];
        with entry{
            for i=0 to i<10 by i+=1 {
                nds.l::append(spawn here ++> node::generic);
            }
            ignore nds[0];
            ignore nds[4];
            take -->;
        }
        report "hi";
    }
"""

async_module = """
walker a {
    can sim1.tester, sim2.tester;
    with entry {
        report sim1.tester();
        report sim2.tester();
    }
}

async walker b {
    can sim1.tester, sim2.tester;
    with entry {
        report sim1.tester();
        report sim2.tester();
    }
}
"""

walker_null_args = """
node c {
    has c1, c2;
}

walker b {
    has anchor b1, b2;
    with entry {
        report b1;
        report b2;
    }
}

walker a {
    has a1;

    with entry {
        a2 = null;
        te = spawn here walker::b(b1 = a1, b2 = a2);
        te = spawn here ++> node::c(c1=a1, c2=a2);
    }
}
"""



from celery import Celery
from celery.app.trace import build_tracer
from celery.app.control import Inspect
from celery.backends.base import DisabledBackend

from jaseci.jsorc.jsorc import JsOrc
from jaseci.jsorc.jsorc_utils import ManifestType
from .tasks import Queue, ScheduledWalker, ScheduledSequence

#################################################
#                   TASK APP                   #
#################################################


@JsOrc.service(name="task", config="TASK_CONFIG")
class TaskService(JsOrc.CommonService):
    ###################################################
    #                   INITIALIZER                   #
    ###################################################

    def __init__(
        self,
        config: dict,
        manifest: dict,
        manifest_type: ManifestType = ManifestType.DEDICATED,
        source: dict = {},
    ):
        self.inspect: Inspect = None
        self.queue: Queue = None
        self.scheduled_walker: ScheduledWalker = None
        self.scheduled_sequence: ScheduledSequence = None

        super().__init__(config, manifest, manifest_type, source)

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self):
        self.app = Celery("celery")
        self.app.conf.update(**self.config)

        # -------------------- TASKS -------------------- #
        (
            self.queue,
            self.scheduled_walker,
            self.scheduled_sequence,
        ) = self.register_tasks(Queue, ScheduledWalker, ScheduledSequence)

        # ------------------ INSPECTOR ------------------ #

        self.inspect = self.app.control.inspect()
        self.ping()

    def post_run(self):
        self.spawn_daemon(
            worker=self.app.Worker(quiet=self.quiet).start,
            scheduler=self.app.Beat(socket_timeout=None, quiet=self.quiet).run,
        )

    def register_tasks(self, *tasks) -> tuple:
        registered = []
        for task in tasks:
            task = self.app.register_task(task())
            task.__trace__ = build_tracer(task.name, task, app=self.app)
            registered.append(task)
        return tuple(registered)

    ###################################################
    #              COMMON GETTER/SETTER               #
    ###################################################

    def get_by_task_id(self, task_id, wait=False, timeout=30):
        task = self.app.AsyncResult(task_id)

        if isinstance(task.backend, DisabledBackend):
            return {
                "status": "DISABLED",
                "result": "result_backend is set to disabled!",
            }

        ret = {"status": task.state}
        if task.ready():
            ret["result"] = task.result
        elif wait:
            ret["status"] = "SUCCESS"
            ret["result"] = task.get(timeout=timeout, disable_sync_subtasks=False)

        return ret

    def inspect_tasks(self):
        return {
            "scheduled": self.inspect.scheduled(),
            "active": self.inspect.active(),
            "reserved": self.inspect.reserved(),
        }

    def ping(self):  # will throw exception
        self.inspect.ping()
        self.app.AsyncResult("").result

    ###################################################
    #                     QUEUING                     #
    ###################################################

    def add_queue(self, wlk, nd, *args):
        return self.queue.delay(wlk.jid, nd.jid, args).task_id

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def failed(self, error):
        super().failed(error)
        self.terminate_daemon("worker", "scheduler")

    # ---------------- PROXY EVENTS ----------------- #

    def on_delete(self):
        self.terminate_daemon("worker", "scheduler")



from jaseci.jac.jsci_vm.op_codes import JsCmp, JsOp, JsType, type_map
from jaseci.jac.jsci_vm.inst_ptr import InstPtr, from_bytes
from jaseci.utils.utils import logger
from base64 import b64decode


class DisAsm(InstPtr):
    def __init__(self):
        InstPtr.__init__(self)
        self._asm = []

    def disassemble(self, bytecode, to_screen=True, log_out=False):
        if type(bytecode) == str:
            bytecode = b64decode(bytecode.encode())
        self._bytecode = bytearray(bytecode)
        try:
            while self._ip < len(self._bytecode):
                op = JsOp(self._bytecode[self._ip])
                if hasattr(self, f"dis_{op.name}"):
                    getattr(self, f"dis_{op.name}")()
                else:
                    self._asm.append([op.name])
                self._ip += 1
            self.print() if to_screen else None
            self.log() if log_out else None
            return self._asm
        except Exception:
            logger.error(f"Disassembly Failed on Following Bytecode: {self._bytecode}")
            self.print() if to_screen else None
            self.log() if log_out else None

    def print(self):
        for i in self._asm:
            print(*i)

    def log(self):
        for i in self._asm:
            logger.info(str(i))

    def op_COMPARE(self):  # noqa
        ctyp = JsCmp(self.offset(1))
        self._asm.append([self.cur_op(), ctyp.name])
        self._ip += 1

    def op_INCREMENT(self):  # noqa
        ityp = JsCmp(self.offset(1))
        self._asm.append([self.cur_op(), ityp.name])
        self._ip += 1

    def dis_LOAD_CONST(self):  # noqa
        typ = JsType(self.offset(1))
        operand2 = self.offset(2)
        val = None
        if typ in [JsType.TYPE]:
            self._asm.append([self.cur_op(), typ.name, JsType(operand2).name])
            self._ip += 2
        elif typ in [JsType.INT]:
            val = from_bytes(type_map[typ], self.offset(3, operand2))
            self._asm.append([self.cur_op(), typ.name, operand2, val])
            self._ip += 2 + operand2
        elif typ in [JsType.STRING]:
            str_len = from_bytes(type_map[JsType.INT], self.offset(3, operand2))
            val = from_bytes(type_map[typ], self.offset(3 + operand2, str_len))
            self._asm.append([self.cur_op(), typ.name, operand2, str_len, val])
            self._ip += 2 + operand2 + str_len
        elif typ in [JsType.FLOAT]:
            val = from_bytes(float, self.offset(2, 8))
            self._asm.append([self.cur_op(), typ.name, val])
            self._ip += 1 + 8
        elif typ in [JsType.BOOL]:
            val = bool(self.offset(2))
            self._asm.append([self.cur_op(), typ.name, val])
            self._ip += 1 + 1

    def dis_LOAD_VAR(self):  # noqa
        name = from_bytes(str, self.offset(2, self.offset(1)))
        self._asm.append([self.cur_op(), self.offset(1), name])
        self._ip += 1 + self.offset(1)

    def dis_CREATE_VAR(self):  # noqa
        name = from_bytes(str, self.offset(2, self.offset(1)))
        self._asm.append([self.cur_op(), self.offset(1), name])
        self._ip += 1 + self.offset(1)

    def dis_DEBUG_INFO(self):  # noqa
        byte_len_l = self.offset(1)
        line = from_bytes(int, self.offset(2, byte_len_l))
        f_offset = byte_len_l + 2
        byte_len_f = self.offset(f_offset)
        jacfile = (
            from_bytes(str, self.offset(f_offset + 1, byte_len_f)) if byte_len_f else 0
        )
        self._asm.append([self.cur_op(), byte_len_l, line, byte_len_f, jacfile])
        self._ip += 2 + byte_len_l + byte_len_f



from jaseci.jac.ir.passes.ir_pass import IrPass


class ParseTreePrunePass(IrPass):
    prune_able = [
        "connect",
        "logical",
        "compare",
        "arithmetic",
        "term",
        "factor",
        "power",
    ]

    cull_able = ["ErrorChar"]

    def enter_node(self, node):
        cull_list = []
        for i in range(len(node.kid)):
            if node.kid[i].name in self.cull_able:
                cull_list.append(node.kid[i])
                continue
            peak = node.kid[i]
            while peak.name in self.prune_able and len(peak.kid) == 1:
                # print("PRUNING:", peak, "from", node.kid, "replacing", peak.kid[0])
                node.kid[i] = peak.kid[0]
                peak = peak.kid[0]
        for i in cull_list:
            node.kid.remove(i)



from unittest import TestCase
from jaseci.utils.utils import TestCaseHelper
from jaseci.cli_tools import jsctl
from click.testing import CliRunner
import json
import os

from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.redis_svc import RedisService


class JsctlTest(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    infer_loc = (
        os.path.dirname(os.path.dirname(os.path.dirname(__file__))) + "/tests/infer.py"
    )

    def setUp(self):
        super().setUp()
        JsOrc.svc("redis", RedisService).clear()

    def call(self, cmd: str):
        res = CliRunner(mix_stderr=False).invoke(jsctl.jsctl, ["-m"] + cmd.split())
        # self.log(res.stdout)
        # self.log(res.stderr)
        # self.log(res.exception)
        return res.stdout

    def jac_call(self, cmd: str):
        res = CliRunner(mix_stderr=False).invoke(jsctl.jac, cmd.split())
        return res.stdout

    def call_cast(self, cmd):
        ret = self.call(cmd)
        # self.log(ret)
        return json.loads(ret)

    def jac_call_cast(self, cmd):
        ret = self.jac_call(cmd)
        # self.log(ret)
        return json.loads(ret)

    def call_split(self, cmd):
        ret = self.call(cmd)
        # self.log(ret)
        return ret.split("\n")

    def jac_call_split(self, cmd):
        ret = self.jac_call(cmd)
        return ret.split("\n")

    def tearDown(self):
        jsctl.reset_state()
        super().tearDown()

    def test_jsctl_extract_tree(self):
        out = jsctl.extract_api_tree()
        self.assertIn("graph", out)
        self.assertIn("walker", out)
        self.assertIn("sentinel", out)
        self.assertIn("alias", out)
        self.assertIn("config", out)
        self.assertNotIn("jadc", out)

    def test_help_screen(self):
        r = self.call("--help")
        self.assertIn("Specify filename", r)
        self.assertIn("Group of `graph` commands", r)
        r = self.call("sentinel --help")
        self.assertIn("Group of `sentinel active` commands", r)

    def test_jsctl_create_graph_mem_only(self):
        ret = self.call(
            f"sentinel register {os.path.dirname(__file__)}/ll.jac -name ll -set_active true"
        )
        self.assertEqual(len(self.call_cast("graph list")), 1)
        self.call("graph create")
        self.call("graph create")
        self.call("graph create")
        self.assertEqual(len(self.call_cast("graph list")), 4)

    def test_jsctl_carry_forward(self):
        ret = self.call(f"actions load local {self.infer_loc}")
        ret = self.call(
            f"sentinel register {os.path.dirname(__file__)}/ll.jac -name ll -set_active true"
        )
        r = self.call("graph create -set_active true")
        r = self.call("walker run init")
        r = self.call("walker run gen_rand_life")
        r = self.call_cast("walker run get_gen_day")
        self.assertGreater(len(r["report"]), 3)

    def test_jsctl_dot(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/ll.jac -name ll -set_active true"
        )
        self.call("walker run init")
        self.call("walker run gen_rand_life")
        r = self.call("graph get -mode dot")
        self.assertIn('"n0" -> "n', r)
        self.assertNotIn('week="', r)

        r = self.call("graph get -mode dot -detailed true")
        self.assertIn('"n0" -> "n', r)
        self.assertIn('id="', r)

    def test_jsctl_aliases(self):
        """Tests that alias mapping api works"""
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )
        gph_id = self.call_cast("graph create")["jid"]
        snt_id = self.call_cast("sentinel list")[0]["jid"]
        self.call(f"alias register s -value {snt_id}")
        self.call(f"alias register g -value {gph_id}")
        self.assertEqual(len(self.call_cast("graph get -nd g")), 1)
        self.call("walker run init -snt s -nd g")
        self.assertEqual(len(self.call_cast("graph get -nd g")), 3)
        self.call("alias clear")
        self.assertEqual(len(self.call_cast("alias list").keys()), 0)

    def test_jsctl_auto_aliases(self):
        """Tests that auto alias mapping api works"""
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )
        aliases = self.call_cast("alias list")
        self.assertGreater(len(aliases), 20)
        self.assertIn("zsb:architype:bot", aliases.keys())
        self.assertIn("zsb:walker:similar_questions", aliases.keys())
        self.assertIn("sentinel:zsb", aliases.keys())
        self.assertIn("zsb:walker:create_nugget", aliases.keys())

    def test_jsctl_auto_aliases_delete(self):
        """Tests that auto removing alias mapping api works"""
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )

        num = len(self.call_cast("alias list"))

        self.call("architype delete zsb:architype:bot")
        self.call("architype delete zsb:walker:similar_questions")
        self.call("architype delete zsb:walker:create_nugget")
        self.assertEqual(num - 3, len(self.call_cast("alias list")))

    def test_jsctl_config_cmds(self):
        """Tests that config commands works"""
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )
        self.call('config set CONFIG_EXAMPLE -value "TEST" -do_check False')
        self.call("config set APPLE -value Grape2 -do_check False")
        self.call("config set APPLE_JSON -value '{\"test\":true}' -do_check False")
        self.call('config set "Banana" -value "Grape" -do_check False')
        self.call('config set "Pear" -value "Banana" -do_check False')
        r = self.call("config get APPLE -do_check False")
        self.assertEqual(r.strip(), "Grape2")
        r = self.call_cast("config list")

        self.assertEqual(len(r), 1)

    def test_jsctl_default_snt_setting(self):
        """Tests that alias mapping api works"""
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/teststest.jac -name zsb "
            "-set_active true"
        )
        snt_id = self.call_cast("sentinel list")[0]["jid"]
        self.call(f"sentinel active set {snt_id}")
        self.call("sentinel active get")
        self.assertEqual(len(self.call_cast("walker list")), 2)

    def test_jsctl_init_auto_called(self):
        """Tests that alias mapping api works"""
        num = len(self.call_cast("graph get"))
        self.assertEqual(num, 3)

    def test_jsctl_master_defaults(self):
        """Tests that alias mapping api works"""
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )
        gph_id = self.call_cast("graph create")["jid"]
        snt_id = self.call_cast("sentinel list")[0]["jid"]
        self.call(f"sentinel active set -snt {snt_id}")
        self.call(f"graph active set -gph {gph_id}")
        self.assertEqual(len(self.call_cast("graph get")), 1)
        self.call("walker run init")
        self.assertEqual(len(self.call_cast("graph get")), 3)

    def test_public_apis_present(self):
        r = self.call("walker --help")
        self.assertIn("summon", r)
        self.assertIn("namespace key", r)

    def test_public_apis_walker_summon_auth(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )
        wjid = self.call_cast("walker spawn create pubinit")["jid"]
        r = self.call_cast(f"walker get {wjid} -mode keys")
        key = r["anyone"]
        r = self.call_cast("alias list")
        nd = r["active:graph"]
        r = self.call_cast(f"walker summon {wjid} -key {key} -nd {nd}")
        self.assertEqual(len(r["report"]), 0)
        key = "aaaaaaaa"
        r = self.call_cast(f"walker summon {wjid} -key {key} -nd {nd}")
        self.assertFalse(r["success"])

    def test_jsctl_import(self):
        self.call(f"actions load local {self.infer_loc}")
        r = self.call_cast(
            f"sentinel register "
            f"{os.path.dirname(__file__)}/ll_base.jac -code_dir "
            f"{os.path.dirname(__file__)} -set_active true"
        )
        r = self.call_cast("walker run init")
        r = self.call_cast("walker run gen_rand_life")
        r = self.call_cast("walker run get_gen_day")
        self.assertGreater(len(r["report"]), 3)

    def test_jsctl_import_filters(self):
        self.call_cast(
            f"sentinel register {os.path.dirname(__file__)}/base.jac -set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_filters1(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/base1.jac -set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_filters2(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/base2.jac -set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_fails_when_incomplete(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/base3.jac -set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertTrue("success" in r.keys())
        self.assertFalse(r["success"])

    def test_jsctl_import_globals(self):
        self.call(
            f"sentinel register "
            f"{os.path.dirname(__file__)}/base4.jac -code_dir {os.path.dirname(__file__)}/ "
            f"-set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_import_recursive(self):
        self.call(
            "sentinel register "
            f"{os.path.dirname(__file__)}/base5.jac -code_dir {os.path.dirname(__file__)}/ "
            "-set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(r["report"][0], "plain")

    def test_jsctl_import_path(self):
        self.call(
            "sentinel register "
            f"{os.path.dirname(__file__)}/base6.jac -code_dir {os.path.dirname(__file__)}/ "
            "-set_active true"
        )
        r = self.call_cast("walker run init")
        self.assertEqual(r["report"][0], "plain")

    def test_jsctl_run_tests(self):
        self.call(f"sentinel register {os.path.dirname(__file__)}/teststest.jac")
        r = self.call_split("sentinel test")
        self.assertTrue(r[0].startswith("Testing assert should be"))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_run_tests_detailed(self):
        self.call(f"sentinel register {os.path.dirname(__file__)}/teststest.jac")
        r = self.call_split("sentinel test -detailed true")
        self.assertEqual(len(r), 33)

    def test_jsctl_multiple_registers_with_globals(self):
        r = self.call_cast(
            f"sentinel register {os.path.dirname(__file__)}/teststest.jir -mode ir"
        )
        r = self.call_cast(
            f"sentinel register {os.path.dirname(__file__)}/teststest.jir -mode ir"
        )
        self.assertEqual(len(r), 1)

    def test_jsctl_run_tests_with_stdout(self):
        self.call(f"sentinel register {os.path.dirname(__file__)}/teststest_stdout.jac")
        r = self.call("sentinel test -detailed true")
        self.assertIn('"stdout": "Some Output\\nSome Output\\n"', r)
        r = r.replace('"stdout": "Some Output\\nSome Output\\n"', "")
        self.assertNotIn("Some Output", r)

    def test_jsctl_jac_build(self):
        if os.path.exists(f"{os.path.dirname(__file__)}/teststest.jir"):
            os.remove(f"{os.path.dirname(__file__)}/teststest.jir")
            self.assertFalse(
                os.path.exists(f"{os.path.dirname(__file__)}/teststest.jir")
            )
        self.call(f"jac build {os.path.dirname(__file__)}/teststest.jac")
        self.assertGreater(
            os.path.getsize(f"{os.path.dirname(__file__)}/teststest.jir"), 20000
        )

    def test_jsctl_jac_build_with_action(self):
        if os.path.exists(f"{os.path.dirname(__file__)}/withaction.jir"):
            os.remove(f"{os.path.dirname(__file__)}/withaction.jir")
            self.assertFalse(
                os.path.exists(f"{os.path.dirname(__file__)}/withaction.jir")
            )
        self.call(f"jac build {os.path.dirname(__file__)}/withaction.jac")
        self.assertGreater(
            os.path.getsize(f"{os.path.dirname(__file__)}/withaction.jir"), 1000
        )
        os.remove(f"{os.path.dirname(__file__)}/withaction.jir")

    def test_jsctl_jac_test(self):
        r = self.call_split(f"jac test {os.path.dirname(__file__)}/teststest.jac")
        self.assertTrue(r[0].startswith("Testing assert should be"))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_jac_test_single(self):
        r = self.call_split(
            f"jac test {os.path.dirname(__file__)}/teststest.jac -single the_second"
        )
        self.log(r)
        self.assertTrue(r[0].startswith("Testing a second test"))
        self.assertTrue(r[2].startswith('  "tests": 1'))
        self.assertTrue(r[5].startswith('  "success": true'))

    def test_jsctl_jac_test_jir(self):
        r = self.call_split(f"jac test {os.path.dirname(__file__)}/teststest.jir")
        self.assertTrue(r[0].startswith("Testing assert should be"))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jac_long_str_build(self):
        r = self.jac_call_cast(f"run {os.path.dirname(__file__)}/longstring.jac")
        self.assertTrue(
            r["report"][0].startswith("Lorem Ipsum is simply dummy text of the")
        )
        self.assertTrue(
            r["report"][0].endswith(
                "Aldus PageMaker including versions of Lorem Ipsum."
            )
        )

    def test_jac_cli_test(self):
        r = self.jac_call_split(f"test {os.path.dirname(__file__)}/teststest.jir")
        self.assertTrue(r[0].startswith("Testing assert should be"))
        self.assertTrue(r[4].startswith('  "tests": 3'))
        self.assertTrue(r[7].startswith('  "success": true'))

    def test_jsctl_jac_disas_jir(self):
        r = self.call(f"jac disas {os.path.dirname(__file__)}/teststest.jir")
        self.assertIn("LOAD_CONST", r)
        self.assertIn("LOAD_VAR", r)

    def test_jsctl_jac_run(self):
        r = self.call_cast(f"jac run {os.path.dirname(__file__)}/teststest.jac")
        self.assertEqual(r["report"], [{}, 4])

    def test_jsctl_jac_run_jir(self):
        r = self.call_cast(f"jac run {os.path.dirname(__file__)}/teststest.jir")
        self.assertEqual(r["report"], [{}, 4])

    def test_jsctl_jac_dot_jir(self):
        r = self.call(f"jac dot {os.path.dirname(__file__)}/teststest.jir")
        self.assertEqual(r, 'strict digraph root {\n    "n0" [ label="n0:root"  ]\n}\n')

    def test_jsctl_jac_run_jir_walk(self):
        r = self.call_cast(
            f"jac run {os.path.dirname(__file__)}/teststest.jir -walk alt_init"
        )
        self.assertEqual(r["report"], [7])

    def test_jsctl_sentinel_set(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/teststest.jac -set_active true"
        )
        self.call(f"sentinel set {os.path.dirname(__file__)}/base.jac")
        r = self.call_cast("walker run init")
        r = self.call_cast("walker run init")
        self.assertEqual(len(r["report"]), 8)

    def test_jsctl_action_list(self):
        r = self.call_cast("actions list")
        self.assertIn("std.out", r)

    def test_jsctl_master_self(self):
        r = self.call_cast("master self")
        self.assertIn("j_type", r.keys())
        a = r["jid"]
        r = self.call_cast("master active get")
        self.assertIn("j_type", r.keys())
        b = r["jid"]
        self.assertEqual(a, b)

    def test_jsctl_graph_can(self):
        self.call(f"actions load local {self.infer_loc}")
        r = self.call(
            f"sentinel register "
            f"{os.path.dirname(__file__)}/graph_can.jac -name gc -set_active true"
        )
        r = self.call("walker run go")
        self.assertEqual(r.split()[0], "2020-01-01T00:00:00")

    def test_jsctl_action_call(self):
        self.call(f"actions load local {self.infer_loc}")
        r = self.call(f"actions call infer.date_now")
        r = json.loads(r)
        import datetime

        date_now = datetime.datetime.utcnow().date().isoformat()
        self.assertEqual(r["result"], date_now)

    def test_jsctl_custom_report(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/glob_imp.jac -set_active true"
        )
        r = self.call_cast("walker run cust_report")
        self.assertEqual(r, {"a": "b"})

    def test_jsctl_custom_report_off(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/glob_imp.jac -set_active true"
        )
        r = self.call_cast("walker run cust_report_neutralize")
        self.assertIn("success", r.keys())

    def test_jsctl_disengage_report(self):
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/glob_imp.jac -set_active true"
        )
        r = self.call_cast("walker run disengage_report")
        self.assertEqual(r, {"a": "b"})

    def test_jsctl_set_global_default_perms(self):
        self.call("object perms default public")
        self.call(
            f"sentinel register {os.path.dirname(__file__)}/zsb.jac -name zsb -set_active true"
        )
        gphs = self.call_cast("graph get -detailed true")
        for i in gphs:
            self.assertEqual(i["j_access"], "public")

    def test_jsctl_bookgen_api_cheatsheet(self):
        r = self.call("booktool cheatsheet")
        self.assertGreater(len(r), 2000)

    def test_jsctl_bookgen_std_library(self):
        r = self.call("booktool stdlib")
        self.assertGreater(len(r), 2000)

    def test_jsctl_bookgen_api_spec(self):
        r = self.call("booktool cheatsheet")
        self.assertGreater(len(r), 2000)

    def test_jsctl_print_detailed_sentinel(self):
        r = self.call_cast(
            f"sentinel register {os.path.dirname(__file__)}/teststest.jir -name test -mode ir"
        )
        r = self.call_cast("object get sentinel:test")
        before = len(r.keys())
        r = self.call_cast("object get sentinel:test -detailed true")
        after = len(r.keys())
        self.assertGreater(before, 4)
        self.assertGreater(after, before)

    def test_jsctl_script(self):
        r = self.call(f"script {os.path.dirname(__file__)}/jsctl_script")
        self.assertEqual(r, "[]\n\n[]\n\n{}\n\n")

    def test_jsctl_script_output(self):
        self.call(f"script {os.path.dirname(__file__)}/jsctl_script -o scr_out")
        with open("scr_out", "r") as f:
            s = [line.rstrip() for line in f]
        if os.path.exists("scr_out"):
            os.remove("scr_out")
        self.assertEqual(
            s,
            [
                "Multi Command Script Output:",
                "Output for sentinel list:",
                "[]",
                "Output for graph list:",
                "[]",
                "Output for alias list:",
                "{}",
            ],
        )


class JsctlTestWithSession(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def call(self, cmd):
        ses = " -f {os.path.dirname(__file__)}/js.session "
        res = CliRunner(mix_stderr=False).invoke(jsctl.jsctl, (ses + cmd).split())
        return res.stdout

    def call_cast(self, cmd):
        ret = self.call(cmd)
        return json.loads(ret)

    def call_split(self, cmd):
        ret = self.call(cmd)
        return ret.split("\n")

    def tearDown(self):
        if os.path.exists(f"{os.path.dirname(__file__)}/js.session"):
            os.remove(f"{os.path.dirname(__file__)}/js.session")
        jsctl.reset_state()
        super().tearDown()

    def test_jsctl_register_with_session(self):
        self.call(f"sentinel register {os.path.dirname(__file__)}/teststest.jac")
        self.call(f"sentinel register {os.path.dirname(__file__)}/teststest.jac")
        r = self.call_cast("object get active:sentinel -detailed true")
        self.assertGreater(len(r["arch_ids"]), 3)



from unittest import TestCase
from unittest.mock import patch
from jaseci.utils.utils import TestCaseHelper
import jaseci.jsorc.live_actions as jla
import jaseci.jsorc.remote_actions as jra
from jaseci.jsorc.live_actions import gen_remote_func_hook


class JacActionsTests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_remote_action_example(self):
        from typing import Union

        @jla.jaseci_action(
            act_group=["use"], aliases=["enc_question"], allow_remote=True
        )
        def question_encode(question: Union[str, list]):
            pass

        @jla.jaseci_action(act_group=["use"], aliases=["enc_answer"], allow_remote=True)
        def answer_encode(answer: Union[str, list], context: Union[str, list] = None):
            pass

        @jla.jaseci_action(act_group=["use"], allow_remote=True)
        def cos_sim_score(q_emb: list, a_emb: list):
            pass

        @jla.jaseci_action(act_group=["use"], aliases=["qa_score"], allow_remote=True)
        def dist_score(q_emb: list, a_emb: list):
            pass

        jra.serv_actions()
        self.assertEqual(
            jra.remote_actions,
            {
                "use.answer_encode": ["answer", "context"],
                "use.cos_sim_score": ["q_emb", "a_emb"],
                "use.dist_score": ["q_emb", "a_emb"],
                "use.enc_answer": ["answer", "context"],
                "use.enc_question": [
                    "question",
                ],
                "use.qa_score": ["q_emb", "a_emb"],
                "use.question_encode": ["question"],
            },
        )

    def test_live_action_globals(self):
        self.assertGreater(len(jla.live_actions), 25)

    @patch("requests.post")
    def test_remote_action_kwargs(self, mock_post):
        remote_action = gen_remote_func_hook(
            url="https://example.com/api/v1/endpoint",
            act_name="example.action",
            param_names=["param1", "param2"],
        )
        payload = {"param1": "value1"}
        remote_action(**payload)

        mock_post.assert_called_once()

        _, kwargs = mock_post.call_args

        assert kwargs["json"] == payload

    def test_setup_decorated_as_startup(self):
        @jla.jaseci_action(act_group=["ex"], allow_remote=True)
        def setup(param: str = ""):
            pass

        app = jra.serv_actions()
        assert len(app.__dict__["router"].__dict__["on_startup"]) == 1



import re
from copy import deepcopy
from typing import Tuple

from celery import Task
from requests import get, post
from requests.exceptions import HTTPError
from jaseci.jsorc.jsorc import JsOrc

DEFAULT_MSG = "Skipping scheduled walker!"


class Queue(Task):
    def run(self, wlk, nd, args):
        hook = JsOrc.hook()

        wlk = hook.get_obj_from_store(wlk)
        wlk._to_await = True

        nd = hook.get_obj_from_store(nd)
        resp = wlk.run(nd, *args)

        # commit to cache first then db instead of `commit()` only
        # this is to support both jsctl and jsserv
        wlk._h.commit_all_cache_sync()
        wlk._h.commit(True)

        wlk.destroy()

        return {"anchor": wlk.anchor_value(), "response": resp}


class ScheduledWalker(Task):
    def get_obj(self, jid):
        return self.hook.get_obj_from_store(jid)

    def run(self, name, ctx, nd=None, snt=None, mst=None):
        self.hook = JsOrc.hook()

        if mst:
            mst = self.get_obj(mst)
        else:
            return f"{DEFAULT_MSG} mst (Master) is required!"

        if mst is None:
            return f"{DEFAULT_MSG} Invalid Master!"

        try:
            if not snt:
                if mst.active_snt_id == "global":
                    global_snt_id = self.hook.get_glob("GLOB_SENTINEL")
                    snt = self.get_obj(global_snt_id)
                elif mst.active_snt_id:
                    snt = self.get_obj(mst.active_snt_id)
            elif snt in mst.alias_map:
                snt = self.get_obj(mst.alias_map[snt])
            else:
                snt = self.get_obj(snt)

            if not snt:
                return f"{DEFAULT_MSG} Invalid Sentinel!"

            if not nd:
                if mst.active_gph_id:
                    nd = self.get_obj(mst.active_gph_id)
            elif nd in mst.alias_map:
                nd = self.get_obj(mst.alias_map[nd])
            else:
                nd = self.get_obj(nd)

            if not nd:
                return f"{DEFAULT_MSG} Invalid Node!"

            return mst.walker_run(name, nd, ctx, ctx, snt, False, False)
        except Exception as e:
            return f"{DEFAULT_MSG} Error occured: {e}"


class ScheduledSequence(Task):
    json_escape = re.compile(r"[^a-zA-Z0-9_]")
    internal = re.compile(r"\(([a-zA-Z0-9_\.\[\]\$\#\@\!]*?)\)")
    full = re.compile(r"^\{\{([a-zA-Z0-9_\.\[\]\$\#\(\)\@\!]*?)\}\}$")
    partial = re.compile(r"\{\{([a-zA-Z0-9_\.\[\]\$\#\(\)\@\!]*?)\}\}")

    def get_deep_value(self, data, keys, default):
        if len(keys) == 0:
            return data

        key = keys.pop(0)
        t = type(data)

        if t is dict and key in data:
            return self.get_deep_value(data[key], keys, default)
        elif t is list and key.isnumeric():
            return self.get_deep_value(data[int(key)], keys, default)
        else:
            return default

    def get_value(self, holder: Tuple, keys: str, default=None):
        while self.internal.search(keys):
            for intern in self.internal.findall(keys):
                keys = keys.replace(
                    "(" + intern + ")", self.get_value(holder, intern, "")
                )

        if keys:
            keys = keys.split(".")
            key = keys.pop(0)
            if key == "#":
                return self.get_deep_value(holder[0], keys, default)
            elif key == "$":
                t = type(holder[1])
                if t is dict or t is list:
                    return self.get_deep_value(holder[1], keys, default)
                else:
                    return holder[1]
            elif key == "@":
                t = type(holder[2])
                if t is dict or t is list:
                    return self.get_deep_value(holder[2], keys, default)
                else:
                    return holder[2]
            elif key == "!":
                return holder[3]
        return default

    def compare(self, condition, expected, actual):
        if condition == "eq":
            return actual == expected
        elif condition == "ne":
            return actual != expected
        elif condition == "gt":
            return actual > expected
        elif condition == "gte":
            return actual >= expected
        elif condition == "lt":
            return actual < expected
        elif condition == "lte":
            return actual <= expected
        elif condition == "regex":
            return re.compile(expected).match(actual)

    def condition(self, holder: Tuple, filter):
        for cons in filter["condition"].keys():
            if not (filter["condition"][cons] is None) and not self.compare(
                cons, filter["condition"][cons], self.get_value(holder, filter["by"])
            ):
                return False
        return True

    def or_condition(self, holder: Tuple, filter):
        for filt in filter:
            if "condition" in filt and self.condition(holder, filt):
                return True
            elif "or" in filt and self.or_condition(holder, filt["or"]):
                return True
            elif "and" in filt and self.and_condition(holder, filt["and"]):
                return True
        return False

    def and_condition(self, holder: Tuple, filter):
        for filt in filter:
            if "condition" in filt and not self.condition(holder, filt):
                return False
            elif "or" in filt and not self.or_condition(holder, filt["or"]):
                return False
            elif "and" in filt and not self.and_condition(holder, filt["and"]):
                return False
        return True

    def deep_replace_str(self, holder: Tuple, data, key):
        matcher = self.full.match(data[key])
        if matcher:
            data[key] = self.get_value(holder, matcher.group(1))
        else:
            for rep in self.partial.findall(data[key]):
                data[key] = data[key].replace(
                    "{{" + rep + "}}", self.get_value(holder, rep, "")
                )

    def deep_replace_dict(self, holder: Tuple, data):
        for key in data.keys():
            if key != "__def_loop__":
                t = type(data[key])
                if t is str:
                    self.deep_replace_str(holder, data, key)
                elif t is dict:
                    self.deep_replace_dict(holder, data[key])

    def save(self, holder: Tuple, req, params):
        if params in req:
            holder[0][self.json_escape.sub("_", req[params])] = holder[1]

    def trigger_interface(self, req: dict):
        master = req.get("master")
        if master is None:
            caller = JsOrc.master()
            trigger_type = "public"
        else:
            caller = JsOrc.hook().get_obj_from_store(master)
            trigger_type = "general"

        api = req.get("api")
        body = req.get("body", {})

        return getattr(caller, f"{trigger_type}_interface_to_api")(body, api)

    def run(self, *args, **kwargs):
        requests = kwargs.get("requests")
        persistence = kwargs.get("persistence", {})
        container = kwargs.get("container", {"current": persistence})
        index = container.get("index", "0")

        if "parent_current" in container:
            container["current"] = container["parent_current"]

        for req in requests:
            try:
                self.deep_replace_dict(
                    (
                        persistence,
                        container["current"],
                        container.get("parent_current", {}),
                        index,
                    ),
                    req,
                )
                self.save((persistence, req), req, "save_req_to")

                method = req["method"].upper()

                if method == "JAC":
                    container["current"] = self.trigger_interface(req)
                else:
                    if method == "POST":
                        response = post(
                            req["api"],
                            json=req.get("body", {}),
                            headers=req.get("header", {}),
                        )
                    elif method == "GET":
                        response = get(req["api"], headers=req.get("header", {}))
                    response.raise_for_status()
                    if "application/json" in response.headers.get("Content-Type"):
                        container["current"] = response.json()
                    else:
                        container["current"] = response.text

                if "__def_loop__" in req:
                    def_loop = req["__def_loop__"]
                    for idx, loop in enumerate(
                        self.get_value(
                            (persistence, container["current"]), def_loop["by"], []
                        )
                    ):
                        if "filter" in def_loop and not self.and_condition(
                            (persistence, loop), def_loop["filter"]
                        ):
                            continue
                        loop_container = {"parent_current": loop, "index": str(idx)}
                        self.run(
                            requests=deepcopy(def_loop["requests"]),
                            persistence=persistence,
                            container=loop_container,
                        )

                self.save((persistence, container["current"]), req, "save_to")

            except Exception as err:
                container["current"] = (
                    {
                        "status": err.response.status_code,
                        "message": err.response.reason,
                    }
                    if isinstance(err, HTTPError)
                    else {
                        "worker_error": str(err),
                    }
                )

                self.save((persistence, container["current"]), req, "save_to")

                if "ignore_error" not in req or not req["ignore_error"]:
                    if "parent_current" in container:
                        raise err
                    break

            if "break" in req and self.and_condition(
                (persistence, container["current"]), req["break"]
            ):
                break

        return persistence



from unittest.mock import MagicMock, Mock

from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.mail_svc import MailService
from jaseci.utils.test_core import CoreTest


class MailLibTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def __init__(self, *args, **kwargs):
        JsOrc.settings("MAIL_CONFIG")["enabled"] = True
        MailService.connect = MagicMock(return_value=Mock())
        super(MailLibTest, self).__init__(*args, **kwargs)

    def setUp(self):
        super().setUp()

    @JsOrc.inject(services=["mail"])
    def test_send_mail(self, mail: MailService):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("mail_test.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "send_mail"}])
        self.assertTrue(ret["success"])
        self.assertTrue(mail.connect.called)
        self.assertEqual(
            mail.app.method_calls[0].args,
            (None, ["jaseci.dev@gmail.com"], "Test Subject", ("Test", "<h1>Test</h1>")),
        )



from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.prome_svc import PrometheusService
from jaseci.jsorc.live_actions import load_action_config
from .actions_optimizer import ActionsOptimizer

import time
import numpy as np


@JsOrc.context("action_manager")
class ActionManager:
    def __init__(self):
        self.benchmark = {
            "jsorc": {"active": False, "requests": {}},
            "actions_optimizer": {"active": False, "requests": {}},
        }
        self.actions_history = {"active": False, "history": []}
        self.actions_calls = {}
        self.system_states = {"active": False, "states": []}
        self.actions_optimizer = ActionsOptimizer(
            benchmark=self.benchmark["actions_optimizer"],
            actions_history=self.actions_history,
            actions_calls=self.actions_calls,
        )

    ###################################################
    #                 ACTION MANAGER                  #
    ###################################################

    def load_action_config(self, config, name):
        """
        Load the config for an action
        """
        return load_action_config(config, name)

    def load_actions(self, name, mode):
        """
        Load an action as local, module or remote.
        """
        # Using module for local
        mode = "module" if mode == "local" else mode

        if mode == "module":
            self.actions_optimizer.load_action_module(name)
        elif mode == "remote":
            self.actions_optimizer.load_action_remote(name)

    def unload_actions(self, name, mode, retire_svc):
        """
        Unload an action
        """
        # We are using module for local
        mode = "module" if mode == "local" else mode
        if mode == "auto":
            res = self.actions_optimizer.unload_action_auto(name)
            if not res[0]:
                return res
            if retire_svc:
                self.retire_uservice(name)
            return res
        elif mode == "module":
            return self.actions_optimizer.unload_action_module(name)
        elif mode == "remote":
            res = self.actions_optimizer.unload_action_remote(name)
            if not res[0]:
                return res
            if retire_svc:
                self.retire_uservice(name)
            return res
        else:
            return (False, f"Unrecognized action mode {mode}.")

    def retire_uservice(self, name):
        """
        Retire a remote microservice for the action.
        """
        self.actions_optimizer.retire_remote(name)

    def get_actions_status(self, name=""):
        """
        Return the status of the action
        """
        return self.actions_optimizer.get_actions_status(name)

    def actions_tracking_start(self):
        """ """
        self.actions_history["active"] = True
        self.actions_history["history"] = [{"ts": time.time()}]
        self.actions_calls.clear()

    def actions_tracking_stop(self):
        """ """
        if not self.actions_history["active"]:
            return []

        self.actions_optimizer.summarize_action_calls()

        return self.actions_history["history"]

    def benchmark_start(self):
        """
        Put JSORC under benchmark mode.
        """
        self.benchmark["jsorc"]["active"] = True
        self.benchmark["jsorc"]["requests"] = {}
        self.benchmark["jsorc"]["start_ts"] = time.time()

    def state_tracking_start(self):
        """
        Ask JSORC to start tracking the state of the system as observed by JSORC on every interval.
        """
        self.system_states = {"active": True, "states": []}

    def state_tracking_stop(self):
        """
        Stop state tracking for JSORC
        """
        ret = self.system_states
        self.system_states = {"active": True, "states": []}
        return ret

    def state_tracking_report(self):
        """
        Return state tracking history so far
        """
        return self.system_states

    def record_system_state(self):
        """
        Record system state
        """
        if self.system_states["active"]:
            ts = int(time.time())
            prom_profile = (
                JsOrc.svc("prome")
                .poke(PrometheusService)
                .info(
                    namespace=self.namespace,
                    exclude_prom=True,
                    timestamp=ts,
                    duration=self.backoff_interval,
                )
            )
            self.system_states["states"].append(
                {
                    "ts": ts,
                    "actions": self.get_actions_status(name=""),
                    "prometheus": prom_profile,
                }
            )

    def benchmark_stop(self, report):
        """
        Stop benchmark mode and report result during the benchmark period
        """
        if not self.benchmark["jsorc"]["active"]:
            return {}

        res = self.benchmark_report()
        self.benchmark["jsorc"]["requests"] = {}
        self.benchmark["jsorc"]["active"] = False

        if report:
            return res
        else:
            return {}

    def benchmark_report(self):
        """
        Summarize benchmark results and report.
        """
        summary = {}
        duration = time.time() - self.benchmark["jsorc"]["start_ts"]
        for request, data in self.benchmark["jsorc"]["requests"].items():
            summary[request] = {}
            all_reqs = []
            for req_name, times in data.items():
                if len(times) == 0:
                    continue
                all_reqs.extend(times)
                summary[request][req_name] = {
                    "throughput": len(times) / duration,
                    "average_latency": sum(times) / len(times) * 1000,
                    "50th_latency": np.percentile(times, 50) * 1000,
                    "90th_latency": np.percentile(times, 90) * 1000,
                    "95th_latency": np.percentile(times, 95) * 1000,
                    "99th_latency": np.percentile(times, 99) * 1000,
                }
            summary[request]["all"] = {
                "throughput": len(all_reqs) / duration,
                "average_latency": sum(all_reqs) / len(all_reqs) * 1000,
                "50th_latency": np.percentile(all_reqs, 50) * 1000,
                "90th_latency": np.percentile(all_reqs, 90) * 1000,
                "95th_latency": np.percentile(all_reqs, 95) * 1000,
                "99th_latency": np.percentile(all_reqs, 99) * 1000,
            }

        return summary

    def record_state(self):
        """
        Record the current state of the system observed by JSORC
        """

    def add_to_benchmark(self, request_type, request, request_time):
        """
        Add requests to benchmark performance tracking
        """
        for bm in self.benchmark.values():
            if request_type not in bm["requests"]:
                bm["requests"][request_type] = {"_default_": []}
            if request_type == "walker_run":
                walker_name = dict(request.data)["name"]
                if walker_name not in bm["requests"][request_type]:
                    bm["requests"][request_type][walker_name] = []
                bm["requests"][request_type][walker_name].append(request_time)
            else:
                bm["requests"][request_type]["_default_"].append(request_time)

    def set_action_policy(self, policy_name, policy_params):
        """
        Set an action optimizer policy
        """
        return self.actions_optimizer.set_action_policy(policy_name, policy_params)

    def get_action_policy(self):
        """
        Get the current action optimization policy
        """
        return self.actions_optimizer.get_action_policy()

    def pre_action_call_hook(self, *args):
        pass

    def post_action_call_hook(self, *args):
        action_name = args[0]
        action_time = args[1]
        if action_name not in self.actions_calls:
            self.actions_calls[action_name] = []

        self.actions_calls[action_name].append(action_time)

    def pre_request_hook(self, *args):
        pass

    def post_request_hook(self, *args):
        request_type = args[0]
        request = args[1]
        request_time = args[2]
        if self.benchmark["jsorc"]["active"]:
            self.add_to_benchmark(request_type, request, request_time)

    def optimize(self, jsorc_interval):
        self.actions_optimizer.run(jsorc_interval)



from base64 import b64decode

from kubernetes import config as kubernetes_config
from kubernetes.client import (
    ApiClient,
    CoreV1Api,
    AppsV1Api,
    RbacAuthorizationV1Api,
    ApiextensionsV1Api,
    AdmissionregistrationV1Api,
    CustomObjectsApi,
)
from kubernetes.client.rest import ApiException

from jaseci.jsorc.jsorc import JsOrc
from jaseci.jsorc.jsorc_utils import ManifestType, placeholder_resolver
from jaseci.utils.utils import logger


@JsOrc.service(name="kube", config="KUBE_CONFIG")
class KubeService(JsOrc.CommonService):
    ###################################################
    #                     BUILDER                     #
    ###################################################

    _no_namespace = [
        "Namespace",
        "ClusterRole",
        "ClusterRoleBinding",
        "CustomResourceDefinition",
        "ValidatingWebhookConfiguration",
    ]

    def run(self):
        self._in_cluster = self.config.get("in_cluster", True)
        if self._in_cluster:
            kubernetes_config.load_incluster_config()
        else:
            kubernetes_config.load_kube_config()

        self.namespace = self.config.get("namespace", "default")
        config = self.config.get("config")

        self.app = ApiClient(config)
        self.core = CoreV1Api(config)
        self.api = AppsV1Api(self.app)
        self.api_ext = ApiextensionsV1Api(self.app)
        self.auth = RbacAuthorizationV1Api(self.app)
        self.reg_api = AdmissionregistrationV1Api(self.app)
        self.custom = CustomObjectsApi(self.app)

        self._cached_namespace = set()

        self.ping()
        self.defaults()

    def defaults(self):
        self.create_apis = {
            "Namespace": self.core.create_namespace,
            "Service": self.core.create_namespaced_service,
            "Deployment": self.api.create_namespaced_deployment,
            "ConfigMap": self.core.create_namespaced_config_map,
            "ServiceAccount": self.core.create_namespaced_service_account,
            "ClusterRole": self.auth.create_cluster_role,
            "ClusterRoleBinding": self.auth.create_cluster_role_binding,
            "Secret": self.core.create_namespaced_secret,
            "PersistentVolumeClaim": (
                self.core.create_namespaced_persistent_volume_claim
            ),
            "DaemonSet": self.api.create_namespaced_daemon_set,
            "StatefulSet": self.api.create_namespaced_stateful_set,
            "CustomResourceDefinition": self.api_ext.create_custom_resource_definition,
            "ValidatingWebhookConfiguration": self.reg_api.create_validating_webhook_configuration,
            "Elasticsearch": lambda namespace, body: self.custom.create_namespaced_custom_object(
                group="elasticsearch.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="elasticsearches",
                body=body,
            ),
            "Kibana": lambda namespace, body: self.custom.create_namespaced_custom_object(
                group="kibana.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="kibanas",
                body=body,
            ),
            "Beat": lambda namespace, body: self.custom.create_namespaced_custom_object(
                group="beat.k8s.elastic.co",
                version="v1beta1",
                namespace=namespace,
                plural="beats",
                body=body,
            ),
        }
        self.patch_apis = {
            "Namespace": self.core.patch_namespace,
            "Service": self.core.patch_namespaced_service,
            "Deployment": self.api.patch_namespaced_deployment,
            "ConfigMap": self.core.patch_namespaced_config_map,
            "ServiceAccount": self.core.patch_namespaced_service_account,
            "ClusterRole": self.auth.patch_cluster_role,
            "ClusterRoleBinding": self.auth.patch_cluster_role_binding,
            "Secret": self.core.patch_namespaced_secret,
            "PersistentVolumeClaim": (
                self.core.patch_namespaced_persistent_volume_claim
            ),
            "DaemonSet": self.api.patch_namespaced_daemon_set,
            "StatefulSet": self.api.patch_namespaced_stateful_set,
            "CustomResourceDefinition": self.api_ext.patch_custom_resource_definition,
            "ValidatingWebhookConfiguration": self.reg_api.patch_validating_webhook_configuration,
            "Elasticsearch": lambda name, namespace, body: self.custom.patch_namespaced_custom_object(
                group="elasticsearch.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="elasticsearches",
                name=name,
                body=body,
            ),
            "Kibana": lambda name, namespace, body: self.custom.patch_namespaced_custom_object(
                group="kibana.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="kibanas",
                name=name,
                body=body,
            ),
            "Beat": lambda name, namespace, body: self.custom.patch_namespaced_custom_object(
                group="beat.k8s.elastic.co",
                version="v1beta1",
                namespace=namespace,
                plural="beats",
                name=name,
                body=body,
            ),
        }
        self.delete_apis = {
            "Namespace": self.core.delete_namespace,
            "Service": self.core.delete_namespaced_service,
            "Deployment": self.api.delete_namespaced_deployment,
            "ConfigMap": self.core.delete_namespaced_config_map,
            "ServiceAccount": self.core.delete_namespaced_service_account,
            "ClusterRole": self.auth.delete_cluster_role,
            "ClusterRoleBinding": self.auth.delete_cluster_role_binding,
            "Secret": self.core.delete_namespaced_secret,
            "PersistentVolumeClaim": (
                self.core.delete_namespaced_persistent_volume_claim
            ),
            "DaemonSet": self.api.delete_namespaced_daemon_set,
            "StatefulSet": self.api.delete_namespaced_stateful_set,
            "CustomResourceDefinition": self.api_ext.delete_custom_resource_definition,
            "ValidatingWebhookConfiguration": self.reg_api.delete_validating_webhook_configuration,
            "Elasticsearch": lambda name, namespace: self.custom.delete_namespaced_custom_object(
                group="elasticsearch.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="elasticsearches",
                name=name,
            ),
            "Kibana": lambda name, namespace: self.custom.delete_namespaced_custom_object(
                group="kibana.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="kibanas",
                name=name,
            ),
            "Beat": lambda name, namespace: self.custom.delete_namespaced_custom_object(
                group="beat.k8s.elastic.co",
                version="v1beta1",
                namespace=namespace,
                plural="beats",
                name=name,
            ),
        }
        self.read_apis = {
            "Namespace": self.core.read_namespace,
            "Service": self.core.read_namespaced_service,
            "Endpoints": self.core.read_namespaced_endpoints,
            "Deployment": self.api.read_namespaced_deployment,
            "ConfigMap": self.core.read_namespaced_config_map,
            "ServiceAccount": self.core.read_namespaced_service_account,
            "ClusterRole": self.auth.read_cluster_role,
            "ClusterRoleBinding": self.auth.read_cluster_role_binding,
            "Secret": self.core.read_namespaced_secret,
            "PersistentVolumeClaim": self.core.read_namespaced_persistent_volume_claim,
            "DaemonSet": self.api.read_namespaced_daemon_set,
            "StatefulSet": self.api.read_namespaced_stateful_set,
            "CustomResourceDefinition": self.api_ext.read_custom_resource_definition,
            "ValidatingWebhookConfiguration": self.reg_api.read_validating_webhook_configuration,
            "Elasticsearch": lambda name, namespace: self.custom.get_namespaced_custom_object(
                group="elasticsearch.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="elasticsearches",
                name=name,
            ),
            "Kibana": lambda name, namespace: self.custom.get_namespaced_custom_object(
                group="kibana.k8s.elastic.co",
                version="v1",
                namespace=namespace,
                plural="kibanas",
                name=name,
            ),
            "Beat": lambda name, namespace: self.custom.get_namespaced_custom_object(
                group="beat.k8s.elastic.co",
                version="v1beta1",
                namespace=namespace,
                plural="beats",
                name=name,
            ),
        }

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def ping(self):
        res = self.app.call_api("/readyz", "GET")
        return res[1] == 200

    def in_cluster(self):
        """
        Check if JSORC/Jaseci is running in a kubernetes cluster
        """
        try:
            return self._in_cluster and self.ping()
        except ApiException as e:
            logger.info(f"Kubernetes cluster environment check failed: {e}")
            return False

    def create(
        self,
        kind: str,
        name: str,
        conf: dict,
        namespace: str,
        log_pref: str = "",
        quiet: bool = False,
    ):
        try:
            quiet or logger.info(
                f"{log_pref} Creating {kind} for `{name}` with namespace: `{namespace}`"
            )
            if kind in self._no_namespace:
                self.create_apis[kind](body=conf)
            else:
                self.create_apis[kind](namespace=namespace, body=conf)
        except ApiException as e:
            quiet or logger.error(
                f"{log_pref} Error creating {kind} for `{name}` with namespace: `{namespace}` -- {e}"
            )

    def patch(
        self,
        kind: str,
        name: str,
        conf: dict,
        namespace: str,
        log_pref: str = "",
        quiet: bool = False,
    ):
        try:
            quiet or logger.info(
                f"{log_pref} Patching {kind} for `{name}` with namespace: `{namespace}`"
            )
            if kind in self._no_namespace:
                self.patch_apis[kind](name=name, body=conf)
            else:
                self.patch_apis[kind](name=name, namespace=namespace, body=conf)
        except ApiException as e:
            quiet or logger.error(
                f"{log_pref} Error patching {kind} for `{name}` with namespace: `{namespace}` -- {e}"
            )

    def read(
        self,
        kind: str,
        name: str,
        namespace: str,
        log_pref: str = "",
        quiet: bool = False,
    ):
        try:
            quiet or logger.info(
                f"{log_pref} Retrieving {kind} for `{name}` with namespace: `{namespace}`"
            )
            if kind in self._no_namespace:
                return self.read_apis[kind](name=name)
            else:
                return self.read_apis[kind](name=name, namespace=namespace)
        except ApiException as e:
            quiet or logger.error(
                f"{log_pref} Error retrieving {kind} for `{name}` with namespace: `{namespace}` -- {e}"
            )
            return e

    def delete(
        self,
        kind: str,
        name: str,
        namespace: str,
        log_pref: str = "",
        quiet: bool = False,
    ):
        try:
            quiet or logger.info(
                f"{log_pref} Deleting {kind} for `{name}` with namespace: `{namespace}`"
            )
            if kind in self._no_namespace:
                return self.delete_apis[kind](name=name)
            else:
                return self.delete_apis[kind](name=name, namespace=namespace)
        except ApiException as e:
            quiet or logger.error(
                f"{log_pref} Error deleting {kind} for `{name}` with namespace: `{namespace}` -- {e}"
            )
            return e

    def is_pod_running(self, name: str, namespace: str):
        try:
            return (
                self.core.list_namespaced_pod(
                    namespace=namespace, label_selector=f"pod={name}"
                )
                .items[0]
                .status.phase
                == "Running"
            )
        except Exception:
            return False

    def get_secret(
        self,
        name: str,
        attr: str,
        namespace: str,
        log_pref: str = "",
        quiet: bool = False,
    ):
        try:
            return b64decode(
                self.core.read_namespaced_secret(
                    name=name,
                    namespace=namespace,
                ).data[attr]
            ).decode()
        except Exception as e:
            quiet or logger.error(
                f"{log_pref} Error getting secret `{attr}` from `{name}` with namespace: `{namespace}` -- {e}"
            )
            return None

    def resolve_manifest(
        self,
        manifest: dict,
        manifest_type: ManifestType = ManifestType.DEDICATED,
        manual_namespace: str = None,
    ) -> dict:
        for kind, confs in manifest.items():
            if kind not in self._no_namespace:
                for conf in confs.values():
                    metadata: dict = conf["metadata"]
                    namespace = metadata.get("namespace", "default")
                    if manifest_type == ManifestType.DEDICATED:
                        namespace = self.namespace
                        metadata["namespace"] = namespace
                    elif manifest_type == ManifestType.MANUAL:
                        namespace = manual_namespace
                        metadata["namespace"] = namespace

                    if namespace and namespace not in self._cached_namespace:
                        res = self.read("Namespace", namespace, None)
                        if hasattr(res, "status") and res.status == 404:
                            self.create(
                                "Namespace",
                                namespace,
                                {
                                    "apiVersion": "v1",
                                    "kind": "Namespace",
                                    "metadata": {
                                        "name": namespace,
                                        "labels": {"name": namespace},
                                    },
                                },
                                None,
                            )
                            # don't add it on cache since create is possible to fail
                        elif (
                            isinstance(res, dict) and "metadata" in res
                        ) or res.metadata:
                            self._cached_namespace.add(namespace)

        placeholder_resolver(manifest, manifest)

        return manifest



from jaseci.utils.test_core import CoreTest


class SentinelApiTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_auto_run_when_synced_to_global(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("hello_world.jac")}],
        )
        self.call(
            self.smast,
            ["global_sentinel_set", {}],
        )
        ret = self.call(
            self.mast,
            [
                "sentinel_active_global",
                {
                    "auto_run": "init",
                    "auto_run_ctx": {"wrld": "jason"},
                    "auto_create_graph": True,
                },
            ],
        )
        num_gphs = len(self.call(self.mast, ["graph_list", {}]))
        self.assertEqual(ret["auto_run_result"]["report"], ["hello jason!"])
        self.assertEqual(num_gphs, 1)

    def test_auto_run_when_synced_to_global_multi(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("hello_world.jac")}],
        )
        self.call(
            self.smast,
            ["global_sentinel_set", {}],
        )
        self.call(
            self.mast,
            [
                "sentinel_active_global",
                {
                    "auto_run": "init",
                    "auto_run_ctx": {"wrld": "jason"},
                    "auto_create_graph": True,
                },
            ],
        )
        ret = self.call(
            self.mast,
            [
                "sentinel_active_global",
                {
                    "auto_run": "init",
                    "auto_run_ctx": {"wrld": "jason"},
                    "auto_create_graph": True,
                },
            ],
        )
        num_gphs = len(self.call(self.mast, ["graph_list", {}]))
        self.assertEqual(ret["auto_run_result"]["report"], ["hello jason!"])
        self.assertEqual(num_gphs, 2)



"""
This module includes code related to hooking Jaseci's Redis to the
core engine.
"""
import json

import jaseci as core_mod
from jaseci.utils.json_handler import JaseciJsonDecoder
from jaseci.jsorc.memory import MemoryHook
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.redis_svc import RedisService


#################################################
#                  REDIS HOOK                   #
#################################################


@JsOrc.repository(name="hook", priority=1)
class RedisHook(MemoryHook):
    def __init__(self, redis: RedisService = None):
        self.redis = JsOrc.svc("redis", RedisService)
        self.red_touch_count = 0

        super().__init__()

    ####################################################
    #        DATASOURCE METHOD (TO BE OVERRIDE)        #
    ####################################################

    # --------------------- OBJ ---------------------- #

    def get_obj_from_store(self, item_id):
        """
        Get item from externally hooked general store by id
        """
        obj = super().get_obj_from_store(item_id)

        if obj is None and self.redis.is_running():
            loaded_obj = self.redis.get(item_id)
            if loaded_obj:
                self.red_touch_count += 1
                jdict = json.loads(loaded_obj, cls=JaseciJsonDecoder)
                j_type = jdict["j_type"]
                j_master = jdict["j_master"]
                class_for_type = self.find_class_and_import(j_type, core_mod)
                ret_obj = class_for_type(h=self, m_id=j_master, auto_save=False)
                ret_obj.json_load(loaded_obj)

                super().commit_obj_to_cache(ret_obj)
                return ret_obj

        return obj

    def has_obj_in_store(self, item_id):
        """
        Checks for object existance in store
        """
        return super().has_obj_in_store(item_id) or (
            self.redis.is_running() and self.redis.exists(item_id)
        )

    # --------------------- GLOB --------------------- #

    def get_glob_from_store(self, name):
        """
        Get global config from externally hooked general store by name
        """
        glob = super().get_glob_from_store(name)

        if glob is None and self.redis.is_running():
            glob = self.redis.hget("global", name)

            if glob:
                self.red_touch_count += 1
                super().commit_glob_to_cache(name, glob)

        return glob

    def has_glob_in_store(self, name):
        """
        Checks for global config existance in store
        """
        return super().has_glob_in_store(name) or (
            self.redis.is_running() and self.redis.hexists("global", name)
        )

    def list_glob_from_store(self):
        """Get list of global config to externally hooked general store"""
        globs = super().list_glob_from_store()

        if not globs and self.redis.is_running():
            globs = self.redis.hkeys("global")

        return globs

    ###################################################
    #   CACHE CONTROL (SHOULD NOT OVERRIDEN ON ORM)   #
    ###################################################

    # -------------------- GLOBS -------------------- #

    def commit_glob_to_cache(self, name, value):
        super().commit_glob_to_cache(name, value)
        if self.redis.is_running():
            self.redis.hset("global", name, value)

    def decommit_glob_from_cache(self, name):
        super().decommit_glob_from_cache(name)

        if self.redis.is_running():
            self.redis.hdel("global", name)

    # --------------------- OBJ --------------------- #

    def commit_obj_to_cache(self, item, all_caches=False):
        super().commit_obj_to_cache(item)

        if all_caches and item._persist and self.redis.is_running():
            self.redis.set(item.jid, item.json(detailed=True))

    def decommit_obj_from_cache(self, item):
        super().decommit_obj_from_cache(item)

        if self.redis.is_running():
            self.redis.delete(item.jid)

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def clear_cache(self):
        if self.redis.is_running():
            self.redis.app.flushdb()

        MemoryHook.__init__(self)


# ----------------------------------------------- #



"""Built in actions for Jaseci"""
from operator import itemgetter
from jaseci.utils.utils import app_logger, json_out
from datetime import datetime
from jaseci.jac.machine.jac_value import jac_wrap_value as jwv
from jaseci.jsorc.live_actions import jaseci_action
from jaseci.utils.utils import master_from_meta
from jaseci.prim.element import Element
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.elastic_svc import Elastic

import sys
import json


@jaseci_action()
def log(*args):
    """Standard built in for printing output to log"""
    result = ""
    for i in args:
        result += json_out(jwv(i))
    app_logger.info(result)
    return result


@jaseci_action()
def out(*args):
    """Standard built in for printing output"""
    args = [json_out(jwv(x)) for x in args]
    print(*args)


@jaseci_action(aliases=["input"])
def js_input(prompt: str = ""):
    """Standard built in for printing output"""
    return input(prompt)


@jaseci_action(aliases=["round"])
def js_round(num: float, digits: int = 0):
    """Standard built in for rounding floats"""
    return round(num, digits)


@jaseci_action()
def err(*args):
    """Standard built in for printing to stderr"""
    args = [json_out(jwv(x)) for x in args]
    print(*args, file=sys.stderr)


@jaseci_action()
def sort_by_col(lst: list, col_num: int, reverse: bool = False):
    """
    Sorts in place list of lists by column
    Param 1 - list
    Param 2 - col number
    Param 3 - boolean as to whether things should be reversed

    Return - Sorted list
    """
    return sorted(lst, key=itemgetter(col_num), reverse=reverse)


# moved to date
@jaseci_action()
def time_now():
    """Get utc date time for now in iso format"""
    return datetime.utcnow().isoformat()


@jaseci_action()
def set_global(name: str, value, meta):
    """
    Set global variable visible to all walkers/users
    Param 1 - name
    Param 2 - value (must be json serializable)
    """
    mast = master_from_meta(meta)
    if not mast.is_master(super_check=True, silent=False):
        return False
    mast.global_set(name, json.dumps(value))
    return json.loads(mast.global_get(name)["value"])


@jaseci_action()
def get_global(name: str, meta):
    """
    Set global variable visible to all walkers/users
    Param 1 - name
    """
    mast = master_from_meta(meta)
    val = mast.global_get(name)["value"]
    if val:
        return json.loads(val)
    else:
        return None


@jaseci_action()
def actload_local(filename: str, meta):
    """
    Load local actions to Jaseci
    """
    mast = master_from_meta(meta)
    if not mast.is_master(super_check=True, silent=True):
        meta["interp"].rt_error("Only super master can load actions.")
        return False
    return mast.actions_load_local(file=filename)["success"]


@jaseci_action()
def actload_remote(url: str, meta):
    """
    Load remote actions to Jaseci
    """
    mast = master_from_meta(meta)
    if not mast.is_master(super_check=True, silent=True):
        meta["interp"].rt_error("Only super master can load actions.")
        return False
    return mast.actions_load_remote(url=url)["success"]


@jaseci_action()
def actload_module(module: str, meta):
    """
    Load module actions to Jaseci
    """
    mast = master_from_meta(meta)
    if not mast.is_master(super_check=True, silent=True):
        meta["interp"].rt_error("Only super master can load actions.")
        return False
    return mast.actions_load_module(mod=module)["success"]


@jaseci_action()
def destroy_global(name: str, meta):
    """Get utc date time for now in iso format"""
    mast = master_from_meta(meta)
    if not mast.is_master(super_check=True, silent=False):
        return False
    return mast.global_delete(name)


@jaseci_action()
def set_perms(obj: Element, mode: str, meta):
    """
    Sets object access mode for any Jaseci object
    Param 1 - target element
    Param 2 - valid permission (public, private, read_only)

    Return - true/false whether successful
    """
    mast = master_from_meta(meta)
    return mast.object_perms_set(obj=obj, mode=mode)["success"]


@jaseci_action()
def get_perms(obj: Element):
    """
    Returns object access mode for any Jaseci object
    Param 1 - target element

    Return - Sorted list
    """
    return obj.j_access


@jaseci_action()
def grant_perms(obj: Element, mast: Element, read_only: bool, meta):
    """
    Grants another user permissions to access a Jaseci object
    Param 1 - target element
    Param 2 - master to be granted permission
    Param 3 - Boolean read_only flag

    Return - Sorted list
    """
    mast = master_from_meta(meta)
    return mast.object_perms_grant(obj=obj, mast=mast, read_only=read_only)["success"]


@jaseci_action()
def revoke_perms(obj: Element, mast: Element, meta):
    """
    Remove permissions for user to access a Jaseci object
    Param 1 - target element
    Param 2 - master to be revoked permission

    Return - Sorted list
    """
    mast = master_from_meta(meta)
    return mast.object_perms_revoke(obj=obj, mast=mast)["success"]


@jaseci_action()
def get_report(meta):
    """
    Get current report so far for walker run
    """
    return meta["interp"].report


@jaseci_action()
def clear_report(meta):
    """
    Clear report so far
    """
    meta["interp"].report = []
    if hasattr(meta["interp"], "save"):
        meta["interp"].save()


@jaseci_action()
def log_activity(
    log: dict = {}, action: str = "", query: str = "", suffix: str = "", meta: dict = {}
):
    elastic = JsOrc.svc("elastic").poke(Elastic)
    activity = elastic.generate_from_meta(meta, log, action)

    return elastic.doc_activity(activity, query, suffix)



"""
Webhook API
"""
from jaseci.extens.api.interface import Interface
from fastapi import HTTPException
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.stripe_svc import StripeService

import stripe as _stripe


class WebhookApi:
    """
    Webhook API
    """

    @Interface.public_api(url_args=["provider"], allowed_methods=["post"])
    def webhook(self, provider: str, _req_ctx: dict = {}, _raw_req_ctx: str = None):
        """Handle webhook logic"""
        req_body = _req_ctx["body"]

        if provider == "stripe":
            stripe_service = JsOrc.svc("stripe", StripeService)
            stripe = stripe_service.poke(_stripe)

            # to be updated
            stripe_service.get_event(_raw_req_ctx, _req_ctx["headers"])

            payload_obj = req_body.get("data").get("object")
            customer_id = payload_obj.get("customer")

            if customer_id:
                customer = stripe.Customer.retrieve(id=customer_id)
                metadata = customer.get("metadata")
            else:
                metadata = payload_obj.get("metadata")

            master_id = metadata.get("master_id")
            master = self._h.get_obj(master_id, master_id)

            node_id = metadata.get("node")
            if not node_id:
                node_id = master.active_gph_id

            node = self._h.get_obj(master_id, node_id)

            global_snt_id = self._h.get_glob("GLOB_SENTINEL")
            global_snt = self._h.get_obj(master_id, global_snt_id)

            payload = {"event": req_body}
            self.seek_committer(master)

            wlk = stripe_service.get_walker(req_body["type"])

            return master.walker_run(
                name=wlk, nd=node, ctx=payload, _req_ctx=_req_ctx, snt=global_snt
            )
        else:
            raise HTTPException(
                status_code=400, detail=str(type + " webhook is not yet supported")
            )



"""Built in actions for Jaseci"""
from jaseci.jsorc.live_actions import jaseci_action
from jaseci.utils.utils import perf_test_start, perf_test_stop

perf_tests = {}


@jaseci_action()
def start_perf_test(name: str = "default"):
    perf_tests[name] = perf_test_start()


@jaseci_action()
def stop_perf_test(name: str = "default"):
    if name not in perf_tests.keys():
        return
    return perf_test_stop(perf_tests[name])



"""
Jac tools api functions as a mixin
"""
from copy import deepcopy
import json
import os

from jaseci.extens.api.interface import Interface
from jaseci.jsorc.jsorc import JsOrc


class JacApi:
    """
    Jac tool APIs
    """

    @Interface.cli_api(cli_args=["file"])
    def jac_build(self, file: str, out: str = "", opt_level: int = 4):
        """
        Command line tooling for building executable jac ir
        """
        filename, dir = self.check_for_file(file)
        if not len(out):
            if file.endswith(".jac"):
                out = file.replace(".jac", ".jir")
            else:
                out = file + ".jir"
        faux = self.faux_master()
        with open(file, "r") as file:
            ret = faux.sentinel_register(
                code=file.read(),
                code_dir=dir,
                name=filename,
                auto_run="",
                opt_level=opt_level,
            )
            if "success" in ret and not ret["success"]:
                return ret
            with open(out, "w") as ofile:
                jir_out = json.dumps(
                    faux.sentinel_get(mode="ir", snt=faux.active_snt())
                )
                ofile.write(jir_out)
                ret = f"Build of {out} complete!"

                return ret

    @Interface.cli_api(cli_args=["file"])
    def jac_disas(self, file: str):
        """
        Command line tooling for print IR for both .jac code files
        and .jir executables
        """
        filename, dir = self.check_for_file(file)
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, "r") as file:
            if is_jir:
                faux.sentinel_register(name=filename)
                ret = faux.sentinel_set(
                    snt=faux.active_snt(), code=file.read(), mode="ir"
                )
                faux.active_snt().print_ir()
            else:
                ret = faux.sentinel_register(
                    code=file.read(), code_dir=dir, name=filename, auto_run=""
                )
                faux.active_snt().print_ir()
            if "success" in ret and not ret["success"]:
                return ret

    @Interface.cli_api(cli_args=["file"])
    def jac_test(
        self,
        file: str,
        single: str = "",
        profiling: bool = False,
        detailed: bool = False,
    ):
        """
        Command line tooling for running all test in both .jac code files
        and .jir executables
        """
        filename, dir = self.check_for_file(file)
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, "r") as file:
            if is_jir:
                faux.sentinel_register(name=filename)
                faux.sentinel_set(snt=faux.active_snt(), code=file.read(), mode="ir")
            else:
                ret = faux.sentinel_register(
                    code=file.read(), code_dir=dir, name=filename, auto_run=""
                )
                if "success" in ret and not ret["success"]:
                    return ret
        return faux.sentinel_test(
            snt=faux.active_snt(), single=single, profiling=profiling, detailed=detailed
        )

    @Interface.cli_api(cli_args=["file"])
    def jac_run(
        self, file: str, walk: str = "init", ctx: dict = {}, profiling: bool = False
    ):
        """
        Command line tooling for running all test in both .jac code files
        and .jir executables
        """
        filename, dir = self.check_for_file(file)
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, "r") as file:
            if is_jir:
                faux.sentinel_register(name=filename)
                faux.sentinel_set(snt=faux.active_snt(), code=file.read(), mode="ir")
            else:
                ret = faux.sentinel_register(
                    code=file.read(), code_dir=dir, name=filename, auto_run=""
                )
                if "success" in ret and not ret["success"]:
                    return ret
        return faux.walker_run(
            name=walk,
            snt=faux.active_snt(),
            nd=faux.active_gph(),
            ctx=ctx,
            profiling=profiling,
        )

    @Interface.cli_api(cli_args=["file"])
    def jac_dot(
        self, file: str, walk: str = "init", ctx: dict = {}, detailed: bool = False
    ):
        """
        Command line tooling for a walker then output graph in both .jac code
        files and .jir executables
        """
        filename, dir = self.check_for_file(file)
        is_jir = file.endswith(".jir")
        faux = self.faux_master()
        with open(file, "r") as file:
            if is_jir:
                faux.sentinel_register(name=filename)
                faux.sentinel_set(snt=faux.active_snt(), code=file.read(), mode="ir")
            else:
                ret = faux.sentinel_register(
                    code=file.read(), name=filename, code_dir=dir, auto_run=""
                )
                if "success" in ret and not ret["success"]:
                    return ret
        faux.walker_run(
            name=walk,
            snt=faux.active_snt(),
            nd=faux.active_gph(),
            ctx=ctx,
            profiling=False,
        )
        return faux.graph_get(nd=faux.active_gph(), mode="dot", detailed=detailed)

    def faux_master(self):
        faux = JsOrc.super_master()
        faux._h.mem["global"] = deepcopy(self._h.mem["global"])
        return faux

    def check_for_file(self, file):
        if not os.path.isfile(file):
            ret = "File does not exsist!"
            return ret, ret
        filename = os.path.basename(file)
        dir = os.path.dirname(os.path.realpath(file))
        return filename, dir



from jaseci.jac.ir.passes.ir_pass import IrPass
from base64 import b64encode


class AstPrunePass(IrPass):
    def enter_node(self, node):
        if hasattr(node, "bytecode") and node.bytecode:
            node.bytecode = b64encode(node.bytecode).decode()
            node.kid = [] if node.name != "inc_assign" else [node.kid[0]]



"""
JSORC APIs
"""
import json
from json import dumps
from time import time
from base64 import b64decode

from jaseci.jsorc.jsorc import JsOrc
from jaseci.jsorc.jsorc_utils import convert_yaml_manifest, ManifestType
from jaseci.utils.utils import logger
from jaseci.extens.svc.kube_svc import KubeService
from jaseci.utils.actions.actions_manager import ActionManager

from jaseci.extens.api.interface import Interface


class JsOrcApi:
    """
    API for managing JsOrc
    """

    @Interface.admin_api()
    def load_yaml(
        self,
        files: list,
        manifest_type: str = "DEDICATED",
        manual_namespace: str = "default",
    ):
        """
        applying list of yaml files without associating to any modules/services
        """

        try:
            kube = JsOrc.svc("kube").poke(KubeService)

            res = {}
            for file in files:
                for kind, confs in kube.resolve_manifest(
                    convert_yaml_manifest(b64decode(file["base64"])),
                    ManifestType[manifest_type],
                    manual_namespace,
                ).items():
                    for name, conf in confs.items():
                        metadata: dict = conf["metadata"]
                        kube.create(
                            kind,
                            metadata["name"],
                            conf,
                            metadata.get("namespace"),
                        )
                        if not res.get(kind):
                            res[kind] = {}
                        res[kind].update({name: conf})
            return res
        except Exception:
            logger.exception("Error loading yaml!")
            return {"message": "load_yaml is not supported on non automated JsOrc!"}

    @Interface.admin_api(cli_args=["service"])
    def apply_yaml(self, service: str, file: list, unsafe_paraphrase: str = ""):
        """
        apply manifest yaml to specific service
        """
        svc = JsOrc.svc(service)

        new_config = {}

        config_version = str(time())

        for kind, confs in convert_yaml_manifest(b64decode(file[0]["base64"])).items():
            for name, conf in confs.items():
                metadata = conf["metadata"]
                labels: dict = metadata.get("labels", {})
                if not labels.get("config_version"):
                    labels["config_version"] = config_version
                    metadata["labels"] = labels

                if not new_config.get(kind):
                    new_config[kind] = {}
                new_config[kind].update({name: conf})

        if unsafe_paraphrase == JsOrc.settings("UNSAFE_PARAPHRASE"):
            new_config["__UNSAFE_PARAPHRASE__"] = unsafe_paraphrase

        self._h.save_glob(svc.source["manifest"], dumps(new_config))
        JsOrc.add_regeneration_queue(service)

        return new_config

    @Interface.admin_api(cli_args=["name"])
    def service_refresh(self, name: str):
        """
        refreshing service's config. If JsOrc is not automated, service will restart else JsOrc will handle the rest
        """

        # will throw exception if not existing
        JsOrc.svc_reset(name)

        return {"success": True}

    @Interface.admin_api(cli_args=["name"])
    def service_call(self, svc: str, attrs: list = []):
        """
        temporary api for retreiving/calling attributes of specific instance.
        """

        svc = JsOrc.svc(svc)

        if not svc:
            return "Service (svc) field is required!"

        if not attrs:
            return "Attributes (attrs) field is required!"

        res = svc
        for at in attrs:
            attr = at.get("attr", False)
            if attr:
                res = getattr(res, attr)

            if at.get("callable", False):
                args = at.get("args", [])
                kwargs = at.get("kwargs", {})
                res = res(*args, **kwargs)

        try:
            # test if json serializable
            json.dumps(res)
            return res
        except Exception:
            return {"error": f"Not JSON Serializable! class {res.__class__.__name__}"}

    @Interface.admin_api(cli_args=["name"])
    def jsorc_actions_load(self, name: str, mode: str):
        """
        Load an action as remote or local mode through JSORC.
        JSORC will load the corresponding module or start a microservice if needed.
        Return the current status of the action.
        """
        action_manager = JsOrc.get("action_manager", ActionManager)
        action_manager.load_actions(name, mode)
        status = action_manager.get_actions_status(name)

        return {"success": True, "action_status": status}

    @Interface.admin_api(cli_args=["name"])
    def jsorc_actions_status(self, name: str):
        """
        Get the current status of an action
        """
        status = JsOrc.get("action_manager", ActionManager).get_actions_status(name)
        return {"success": True, "action_status": status}

    @Interface.admin_api(cli_args=["name"])
    def jsorc_actions_unload(
        self, name: str, mode: str = "auto", retire_svc: bool = True
    ):
        """
        Unload an action through JSORC.
        If retire_svc is set to True (true by default), it will also retire the corresponding microservice.
        """
        res = JsOrc.get("action_manager", ActionManager).unload_actions(
            name, mode, retire_svc
        )
        return {"success": res[0], "message": res[1]}

    @Interface.admin_api(cli_args=["config", "name"])
    def jsorc_actions_config(self, config: str, name: str):
        """
        Loading the config of an action module
        config: name of the ai kit package (e.g. jac_nlp.config, jac_vision.config)
        name: name of the action module (e.g. use_enc, bi_enc)
        """
        res = JsOrc.get("action_manager", ActionManager).load_action_config(
            config, name
        )
        return {"success": res}

    @Interface.admin_api()
    def jsorc_trackact_start(self):
        """ "
        Instruct JSORC to start tracking any changes in actions state
        """
        JsOrc.get("action_manager", ActionManager).actions_tracking_start()
        return {"success": True}

    @Interface.admin_api()
    def jsorc_trackact_stop(self):
        """ "
        Instruct JSORC to start tracking any changes in actions state
        """
        return JsOrc.get("action_manager", ActionManager).actions_tracking_stop()

    @Interface.admin_api()
    def jsorc_benchmark_start(self):
        """
        Tell JSORC to start collecting request performance metrics
        """
        JsOrc.get("action_manager", ActionManager).benchmark_start()
        return {"success": True}

    @Interface.admin_api()
    def jsorc_benchmark_report(self):
        """
        Report the collected request performance metrics of the currently ongoing benchmark
        """
        return JsOrc.get("action_manager", ActionManager).benchmark_report()

    @Interface.admin_api()
    def jsorc_benchmark_stop(self, report: bool = True):
        """
        End the benchmark period and report performance metrics
        """
        return JsOrc.get("action_manager", ActionManager).benchmark_stop(report)

    @Interface.admin_api()
    def jsorc_tracksys_start(self):
        """
        Ask JSORC to start tracking the state of the system as observed by JSORC on every interval.
        """
        JsOrc.get("action_manager", ActionManager).state_tracking_start()
        return {"success": True}

    @Interface.admin_api()
    def jsorc_tracksys_report(self):
        """
        Report the tracked system states so far
        """
        return JsOrc.get("action_manager", ActionManager).state_tracking_report()

    @Interface.admin_api()
    def jsorc_tracksys_stop(self):
        """
        Stop state tracking for JSORC
        """
        return JsOrc.get("action_manager", ActionManager).state_tracking_stop()

    @Interface.admin_api()
    def jsorc_actionpolicy_set(self, policy_name: str, policy_params: dict = {}):
        """
        Set an action optimization policy for JSORC
        """
        res = JsOrc.get("action_manager", ActionManager).set_action_policy(
            policy_name, policy_params
        )
        if res is True:
            return {
                "success": True,
                "message": f"Action optimization policy configured as {policy_name} with params {policy_params}",
            }
        else:
            return {"success": False, "message": res}

    @Interface.admin_api()
    def jsorc_actionpolicy_get(self):
        """
        Get the current active action optimization policy
        """
        policy = JsOrc.get("action_manager", ActionManager).get_action_policy()
        return {"success": True, "policy": policy}

    @Interface.admin_api()
    def jsorc_loadtest(self, test: str, experiment: str = "", mem: int = 0):
        """
        load test API. overwritten in jaseci_serv
        """
        pass



"""
User API
"""
from jaseci.extens.api.interface import Interface
from jaseci.jsorc.jsorc import JsOrc


class UserApi:
    """
    User APIs for creating users (some functions should be override downstream)

    These User APIs enable the creation and management of users on a Jaseci machine.
    The creation of a user in this context is synonymous to the creation of a master
    Jaseci object. These APIs are particularly useful when running a Jaseci server
    or cluster in contrast to running JSCTL on the command line. Upon executing JSCTL
    a dummy admin user (super_master) is created and all state is dumped to a session
    file, though any users created during a JSCTL session will indeed be created as
    part of that session's state.
    """

    @Interface.public_api(cli_args=["name"])
    def user_create(
        self,
        name: str,
        password: str = "",
        global_init: str = "",
        global_init_ctx: dict = {},
        other_fields: dict = {},
    ):
        """
        Create a new user (master object)

        This API is used to create users and optionally set them up with a graph and
        related initialization. In the context of
        JSCTL, any name is sufficient and no additional information is required.
        However, for Jaseci serving (whether it be the official Jaseci server, or a
        custom overloaded server) additional fields are required and should be added
        to the other_fields parameter as per the specifics of the encapsulating server
        requirements. In the case of the official Jaseci server, the name field must
        be a valid email, and a password field must be passed through other_fields.
        A number of other optional parameters can also be passed through other_feilds.
        \\vspace{3mm}\\par
        This single API call can also be used to fully set up and initialize a user
        by leveraging the global_init parameter. When set, this parameter attaches the
        user to the global sentinel, creates a new graph for the user, sets it as the
        active graph, then runs an initialization walker on the root node of this new
        graph. The initialization walker is identified by the name assigned to
        global_init. The default empty string assigned to global_init indicates this
        global setup should not be run.

        :param name: The user name to create. For Jaseci server this must be a valid
            email address.
        :param global_init: The name of an initialization walker. When set the user is
            linked to the global sentinel and the walker is run on a new active graph
            created for the user.
        :param global_init_ctx: Context to preload for the initialization walker
        :param other_fields: This parameter is used for additional fields required for
            overloaded interfaces. This parameter is not used in JSCTL, but is used
            by Jaseci server for the additional parameters of password, is_activated,
            and is_superuser.
        """
        ret = {}
        mast = self.user_creator(name, password, other_fields)
        if type(mast) is dict:  # in case of upstream error
            return mast
        ret["user"] = mast.serialize()
        self.seek_committer(mast)
        if len(global_init):
            ret["global_init"] = self.user_global_init(
                mast, global_init, global_init_ctx
            )
        ret["success"] = True
        return ret

    @Interface.admin_api(cli_args=["name"])
    def user_delete(self, name: str):
        """
        Delete new user (master object)

        This API is used to delete a user account.

        :param name: The user name to delete. For Jaseci server this must
        be a valid email address.

        """
        ret = {}
        ret["success"] = self.user_destroyer(name)
        if not ret["success"]:
            ret["status_code"] = 400
        return ret

    @Interface.private_api(cli_args=["name"])
    def user_deleteself(self):
        """
        Delete self (master object)

        This API is used to delete a user account.
        """
        ret = {}
        ret["success"] = self.user_destroyer(self.name)
        if not ret["success"]:
            ret["status_code"] = 400
        return ret

    def user_creator(self, name, password: str = "", other_fields: dict = {}):
        """
        Abstraction for user creation for elegant overriding
        """

        return JsOrc.master(h=self._h, name=name)

    def superuser_creator(self, name, password: str = "", other_fields: dict = {}):
        """
        Abstraction for super user creation for elegant overriding
        """

        return JsOrc.super_master(h=self._h, name=name)

    def user_global_init(
        self,
        mast,
        global_init: str = "",
        global_init_ctx: dict = {},
    ):
        """
        Create a master instance and return root node master object

        other_fields used for additional fields for overloaded interfaces
        (i.e., Dango interface)
        """
        return mast.sentinel_active_global(
            auto_run=global_init,
            auto_run_ctx=global_init_ctx,
            auto_create_graph=True,
        )

    def user_destroyer(self, name: str):
        """
        Permanently delete master with given id
        """
        pass



from jaseci.jac.ir.passes import IrPass
from jaseci.jac.jsci_vm.op_codes import JsCmp, JsOp, JsType
from struct import pack

from jaseci.utils.utils import parse_str_token


def byte_length(val):
    if type(val) == str:
        return len(bytes(val, "unicode_escape"))
    else:
        return (val.bit_length() + 7) // 8


def to_bytes(val):
    if type(val) == str:
        return bytes(val, "unicode_escape")
    elif type(val) == float:
        return pack("d", val)
    elif type(val) == int:
        return val.to_bytes(byte_length(val), "little")


def has_bytecode(node):
    if not hasattr(node, "bytecode"):
        return False
    return True


def is_bytecode_complete(node):
    for i in node.kid:
        if not i.is_terminal() and not has_bytecode(i) and i.name != "cmp_op":
            return False
    return True


class CodeGenPass(IrPass):
    def __init__(self, debug_info=True, **kwargs):
        super().__init__(**kwargs)
        self.debug_info = debug_info
        self.cur_loc = None
        self.create_var_mode = 0

    def emit(self, node, *items):
        if not has_bytecode(node):
            node.bytecode = bytearray()
        node_loc = [node.loc[0], node.loc[2]]
        if self.debug_info and self.cur_loc != node_loc:
            debug_inst = [
                JsOp.DEBUG_INFO,
                byte_length(node_loc[0]),
                to_bytes(node_loc[0]),
            ]
            if not self.cur_loc or self.cur_loc[1] != node_loc[1]:
                debug_inst += [byte_length(node_loc[1]), to_bytes(node_loc[1])]
            else:
                debug_inst += [0]
            self.cur_loc = node_loc
            items = debug_inst + list(items)
        for i in items:
            if type(i) in [bytes, bytearray]:
                node.bytecode += bytearray(i)
            elif type(i) is str:
                node.bytecode += bytearray(i, "utf-8")
            else:
                node.bytecode.append(i)

    def enter_node(self, node):
        # print("entering", node)
        if hasattr(self, f"enter_{node.name}"):
            getattr(self, f"enter_{node.name}")(node)

    def exit_node(self, node):
        # print("exiting", node)
        if hasattr(self, f"exit_{node.name}"):
            getattr(self, f"exit_{node.name}")(node)
        if hasattr(node, "_create_flag"):
            self.create_var_mode -= 1

    def enter_expression(self, node):
        kid = node.kid
        if len(kid) > 1 and kid[-1].name == "assignment":
            kid[-1]._create_flag = True
            self.create_var_mode += 1

    def exit_expression(self, node):  # TODO: Incomplete
        kid = node.kid
        if is_bytecode_complete(node):
            for i in reversed(node.kid):
                if has_bytecode(i):
                    self.emit(node, i.bytecode)
            if len(kid) > 1 and kid[1].name == "assignment":
                self.emit(node, JsOp.ASSIGN)
            elif len(kid) > 1 and kid[1].name == "copy_assign":
                self.emit(node, JsOp.COPY_FIELDS)
            elif len(kid) > 1 and kid[1].name == "inc_assign":
                self.emit(node, JsOp.INCREMENT, JsCmp[kid[1].kid[0].name])

    def exit_assignment(self, node):
        if is_bytecode_complete(node):
            if has_bytecode(node.kid[-1]):
                self.emit(node, node.kid[-1].bytecode)

    def exit_copy_assign(self, node):
        if is_bytecode_complete(node):
            if has_bytecode(node.kid[-1]):
                self.emit(node, node.kid[-1].bytecode)

    def exit_inc_assign(self, node):
        if is_bytecode_complete(node):
            if has_bytecode(node.kid[-1]):
                self.emit(node, node.kid[-1].bytecode)

    def exit_logical(self, node):
        if is_bytecode_complete(node):
            for i in reversed(node.kid):
                if has_bytecode(i):
                    self.emit(node, i.bytecode)
            for i in node.kid:
                if i.name == "KW_AND":
                    self.emit(node, JsOp.AND)
                elif i.name == "KW_OR":
                    self.emit(node, JsOp.OR)

    def exit_compare(self, node):
        if is_bytecode_complete(node):
            for i in reversed(node.kid):
                if has_bytecode(i):
                    self.emit(node, i.bytecode)
            for i in node.kid:
                if i.name == "NOT":
                    self.emit(node, JsOp.COMPARE, JsCmp.NOT)
                elif i.name == "cmp_op" and i.kid[0].name == "EE":
                    self.emit(node, JsOp.COMPARE, JsCmp.EE)
                elif i.name == "cmp_op" and i.kid[0].name == "LT":
                    self.emit(node, JsOp.COMPARE, JsCmp.LT)
                elif i.name == "cmp_op" and i.kid[0].name == "GT":
                    self.emit(node, JsOp.COMPARE, JsCmp.GT)
                elif i.name == "cmp_op" and i.kid[0].name == "LTE":
                    self.emit(node, JsOp.COMPARE, JsCmp.LTE)
                elif i.name == "cmp_op" and i.kid[0].name == "GTE":
                    self.emit(node, JsOp.COMPARE, JsCmp.GTE)
                elif i.name == "cmp_op" and i.kid[0].name == "NE":
                    self.emit(node, JsOp.COMPARE, JsCmp.NE)
                elif i.name == "cmp_op" and i.kid[0].name == "KW_IN":
                    self.emit(node, JsOp.COMPARE, JsCmp.IN)
                elif i.name == "cmp_op" and i.kid[0].name == "nin":
                    self.emit(node, JsOp.COMPARE, JsCmp.NIN)

    def exit_arithmetic(self, node):
        if is_bytecode_complete(node):
            for i in reversed(node.kid):
                if has_bytecode(i):
                    self.emit(node, i.bytecode)
            for i in node.kid:
                if i.name == "PLUS":
                    self.emit(node, JsOp.ADD)
                elif i.name == "MINUS":
                    self.emit(node, JsOp.SUBTRACT)

    def exit_term(self, node):
        if is_bytecode_complete(node):
            for i in reversed(node.kid):
                if has_bytecode(i):
                    self.emit(node, i.bytecode)
            for i in node.kid:
                if i.name == "STAR_MUL":
                    self.emit(node, JsOp.MULTIPLY)
                elif i.name == "DIV":
                    self.emit(node, JsOp.DIVIDE)
                elif i.name == "MOD":
                    self.emit(node, JsOp.MODULO)

    def exit_factor(self, node):
        if is_bytecode_complete(node):
            self.emit(node, node.kid[-1].bytecode)
            if node.kid[0].name == "MINUS":
                self.emit(node, JsOp.NEGATE)

    def exit_power(self, node):
        if is_bytecode_complete(node):
            for i in reversed(node.kid):
                if has_bytecode(i):
                    self.emit(node, i.bytecode)
            for i in node.kid:
                if i.name == "POW":
                    self.emit(node, JsOp.POWER)

    def exit_atom(self, node):  # TODO: Incomplete
        kid = node.kid
        if kid[0].name == "INT":
            val = int(kid[0].token_text())
            self.emit(
                node, JsOp.LOAD_CONST, JsType.INT, byte_length(val), to_bytes(val)
            )
        elif kid[0].name == "FLOAT":
            val = float(kid[0].token_text())
            self.emit(node, JsOp.LOAD_CONST, JsType.FLOAT, to_bytes(val))
        elif kid[0].name == "multistring":
            val = ""
            for i in kid[0].kid:
                val += parse_str_token(i.token_text())
            str_len = byte_length(val)
            self.emit(
                node,
                JsOp.LOAD_CONST,
                JsType.STRING,
                byte_length(str_len),
                to_bytes(str_len),
            )
            if str_len > 0:
                self.emit(node, to_bytes(val))
        elif kid[0].name == "BOOL":
            val = int(kid[0].token_text() == "true")
            self.emit(node, JsOp.LOAD_CONST, JsType.BOOL, val)
        elif kid[0].name == "NULL":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.NULL)
        elif kid[0].name == "NAME":
            name = kid[0].token_text()
            if self.create_var_mode:
                self.emit(node, JsOp.CREATE_VAR, byte_length(name), to_bytes(name))
            else:
                self.emit(node, JsOp.LOAD_VAR, byte_length(name), to_bytes(name))

    def exit_any_type(self, node):
        kid = node.kid
        if kid[0].name == "TYP_STRING":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.STRING)
        elif kid[0].name == "TYP_INT":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.INT)
        elif kid[0].name == "TYP_FLOAT":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.FLOAT)
        elif kid[0].name == "TYP_LIST":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.LIST)
        elif kid[0].name == "TYP_DICT":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.DICT)
        elif kid[0].name == "TYP_BOOL":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.BOOL)
        elif kid[0].name == "KW_NODE":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.NODE)
        elif kid[0].name == "KW_EDGE":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.EDGE)
        elif kid[0].name == "KW_TYPE":
            self.emit(node, JsOp.LOAD_CONST, JsType.TYPE, JsType.TYPE)



"""
Alias api as a mixin
"""
from jaseci.extens.api.interface import Interface


class AliasAPI:
    """
    Alias APIs for creating nicknames for UUIDs and other long strings

    The alias set of APIs provide a set of `alias' management functions for
    creating and managing aliases for long strings such as UUIDs. If an alias'
    name is used as a parameter value in any API call, that parameter will see
    the alias' value instead. Given that references to all sentinels, walkers,
    nodes, etc. utilize UUIDs, it becomes quite useful to create pneumonic
    names for them. Also, when registering   sentinels, walkers, architype
    handy aliases are automatically generated. These generated aliases can
    then be managed using the alias APIs. Keep in mind that whenever an alias
    is created, all parameter values submitted to any API with the alias name
    will be replaced internally by its value. If you get in a bind, simply use
    the clear or delete alias APIs.
    """

    def __init__(self):
        self.alias_map = {}

    @Interface.private_api(cli_args=["name"])
    def alias_register(self, name: str, value: str):
        """Create string to string alias mapping that caller can use.

        Either create new alias string to string mappings or replace
        an existing mappings of a given alias name. Once registered the
        alias mapping is instantly active.

        Args:
            name (str): The name for the alias created by caller.
            value (str): The value for that name to map to (i.e., UUID)

        Returns:
            json: Fields include
                'response': Message of mapping that was created
        """
        self.alias_map[name] = value
        self.save()
        return {"response": f"Alias from '{name}' to '{value}' set!"}

    @Interface.private_api()
    def alias_list(self):
        """List all string to string alias that caller can use.

        Returns dictionary object of name to value mappings currently active.
        This API is quite useful to track not only the aliases the caller
        creates, but also the aliases automatically created as various Jaseci
        objects (walkers, architypes, sentinels, etc.) are created, changed,
        or destroyed.

        Returns:
            dictionary: Dictionary of active mappings
                'name': 'value'
                ...
        """
        return self.alias_map

    @Interface.private_api(cli_args=["name"])
    def alias_delete(self, name: str):
        """Delete an active string to string alias mapping.

        Removes a specific alias by its name. Only the alias is removed no
        actual objects are affected. Future uses of this name will not be
        mapped.

        Args:
            name (str): The name for the alias to be removed from caller.

        Returns:
            dictionary: Fields include
                'response': Message of success/failure to remove alias
                'success': True/False based on delete actually happening
        """
        if name in self.alias_map.keys():
            del self.alias_map[name]
            self.save()
            return {"response": f"Alias {name} successfully deleted", "success": True}
        else:
            return {"response": f"Alias {name} not present", "success": False}

    @Interface.private_api()
    def alias_clear(self):
        """Remove all string to string alias that client can use.

        Removes a all aliases. No actual objects are affected. Aliases will
        continue to be automatically generated when creating other Jaseci
        objects.

        Returns:
            dictionary: Fields include
                'response': Message of number of alias removed
                'removed': Number of aliases removed
        """
        n = len(self.alias_map.keys())
        self.alias_map = {}
        self.save()
        return {"response": f"All {n} aliases deleted", "removed": n}

    def extract_snt_aliases(self, snt):
        """
        Extract and register all aliases from sentinel
        """
        self.alias_register(f"sentinel:{snt.name}", snt.jid)
        for i in snt.arch_ids.obj_list():
            self.extract_arch_aliases(snt, i)
        self.save()

    def extract_arch_aliases(self, snt, arch):
        """
        Extract and register all aliases from architype
        """
        if arch.kind == "walker":
            self.alias_register(f"{snt.name}:walker:{arch.name}", arch.jid)
        else:
            self.alias_register(f"{snt.name}:architype:{arch.name}", arch.jid)

    def remove_snt_aliases(self, snt):
        """
        Extract and register all aliases from sentinel
        """
        self.alias_delete(f"sentinel:{snt.name}")
        for i in snt.arch_ids.obj_list():
            self.remove_arch_aliases(snt, i)
        self.save()

    def remove_arch_aliases(self, snt, arch):
        """
        Extract and register all aliases from architype
        """
        if arch.kind == "walker":
            self.alias_delete(f"{snt.name}:walker:{arch.name}")
        else:
            self.alias_delete(f"{snt.name}:architype:{arch.name}")



from jaseci.utils.test_core import CoreTest, jac_testcase


class RegexTest(CoreTest):
    """UnitTest for Regex Module"""

    fixture_src = __file__

    @jac_testcase("regex.jac", "findall_test")
    def test_findall(self, ret):
        ret1 = [ret["report"][0][0]]
        ret2 = [ret["report"][0][1]]
        self.assertEqual(ret1, [("width", "20")])
        self.assertEqual(ret2, [("height", "10")])

    @jac_testcase("regex.jac", "search_test")
    def test_search(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, [{"span": (5, 7), "match": "ai"}])

    @jac_testcase("regex.jac", "match_test")
    def test_match(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, [{"span": (0, 1), "match": "T"}])

    @jac_testcase("regex.jac", "fullmatch_test")
    def test_fullmatch(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, [{"span": (0, 16), "match": "THEQUICKBROWNFOX"}])

    @jac_testcase("regex.jac", "split_test")
    def test_split(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, [["Words", "words", "words", ""]])

    @jac_testcase("regex.jac", "finditer_test")
    def test_finditer(self, ret):
        ret = ret["report"]
        self.assertEqual(
            ret,
            [[{"span": (0, 4), "match": "Blue"}, {"span": (13, 17), "match": "Blue"}]],
        )

    @jac_testcase("regex.jac", "sub_test")
    def test_sub(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, ["Account Number - NN, Amount - NN.NN"])

    @jac_testcase("regex.jac", "subn_test")
    def test_subn(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, [["Account Number - NN, Amount - NN.NN", 3]])

    @jac_testcase("regex.jac", "escape_test")
    def test_escape(self, ret):
        ret = ret["report"]
        self.assertEqual(ret, ["https://www\\.jaseci\\.org"])

    @jac_testcase("regex.jac", "purge_test")
    def test_purge(self, ret):
        self.assertEqual(ret["success"], True)



import stripe
from jaseci.jsorc.jsorc import JsOrc

#################################################
#                  STRIPE APP                   #
#################################################


@JsOrc.service("stripe", config="STRIPE_CONFIG")
class StripeService(JsOrc.CommonService):
    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self):
        self.app = stripe
        if not self.config.get("api_key"):
            raise Exception("api_key is required!")
        self.app.api_key = self.config.get("api_key")

        if not self.config.get("webhook_key"):
            raise Exception("webhook_key is required!")
        self.webhook_key = self.config.get("webhook_key")

        self.fallback_walker = self.config.get("fallback_walker")
        if not self.fallback_walker:
            raise Exception("fallback_walker is required!")

        self.event_walker = self.config.get("event_walker", {})

    ####################################################
    #               COMMON GETTER/SETTER               #
    ####################################################

    def get_walker(self, event):
        return self.event_walker.get(event) or self.fallback_walker

    def get_event(self, body, headers):
        return stripe.Webhook.construct_event(
            body,
            headers.get("Stripe-Signature"),
            self.webhook_key,
        )

    ####################################################
    #                    OVERRIDDEN                    #
    ####################################################

    def reset(self, hook, start=True):
        stripe.api_key = None
        super().reset(hook, start)



"""Built in actions for Jaseci"""
from jaseci.jsorc.live_actions import jaseci_action
from fastapi import HTTPException
import base64
import requests
import urllib
import validators


@jaseci_action()
def is_valid(url: str = ""):
    """
    Checks that a string is a valid url in its format
    Param 1 - url string

    Return - True if valid, False if not
    """
    try:
        url_response = validators.url(url)
        if isinstance(url_response, validators.ValidationFailure):
            return False
    except:
        return False

    return True


@jaseci_action()
def ping(url: str = ""):
    """
    Make sure that a url returns an error code in the 200s when pinged upon success
    Param 1 - url string

    Return - True if error code is in 200s, False if not
    """
    try:
        response = requests.get(url)
    except:
        return False

    return response.status_code in range(200, 299)


@jaseci_action()
def download_text(url: str = ""):
    """
    Downloaded html from a url
    Param 1 - url string

    Return - string containing html
    """
    try:
        response = urllib.request.urlopen(url)
    except:
        raise HTTPException(status_code=404, detail="Invalid URL")

    html_bytes = response.read()
    html_str = html_bytes.decode("UTF-8")
    print(len(html_str))
    return html_str


@jaseci_action()
def download_b64(url: str = ""):
    """
    Downlaod the get response from a url and return it in base64
    Param 1 - url string

    Return - base64 string representing the get response
    """
    try:
        response = requests.get(url)
    except:
        raise HTTPException(status_code=404, detail="Invalid URL")

    response_bytes = response.content
    response_b64_bytes = base64.b64encode(response_bytes)
    response_b64 = response_b64_bytes.decode("ascii")
    return response_b64



import os
from antlr4 import ParseTreeListener
from antlr4.error.ErrorListener import ErrorListener
from jaseci.utils.utils import logger, parse_str_token
from jaseci.jac.jac_parse.jacParser import jacParser, ParseTreeWalker
from jaseci.jac.jac_parse.jacLexer import jacLexer
from antlr4 import InputStream, CommonTokenStream
from jaseci.jac.ir.ast import Ast


class JacAstBuilder:
    """
    Jac Code to AST Tree
    """

    _ast_head_map = {}

    def __init__(
        self,
        mod_name,
        mod_dir="./",
        jac_text=None,
        start_rule="start",
    ):
        self.root = Ast(mod_name)
        self._parse_errors = []
        self._start_rule = start_rule
        self._mod_dir = mod_dir
        self.dependencies = []
        if jac_text:
            self.jac_code_to_ast(jac_text)

    def jac_code_to_ast(self, jac_str):
        """Parse language and build ast from string"""
        JacAstBuilder._ast_head_map[self._mod_dir + self.root.loc[2]] = self
        input_stream = InputStream(jac_str)
        lexer = jacLexer(input_stream)
        stream = CommonTokenStream(lexer)
        errors = JacTreeError(self)
        parser = jacParser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(errors)
        tree = getattr(parser, self._start_rule)()
        builder = JacTreeBuilder(builder=self, code=jac_str)
        walker = ParseTreeWalker()
        walker.walk(builder, tree)

        if self._parse_errors:
            logger.error(str(f"Parse errors encountered - {self}"))


class JacTreeBuilder(ParseTreeListener):
    """Converter class from Antlr trees to Jaseci Tree"""

    def __init__(self, builder, code):
        self.builder = builder
        self.code = code
        self.node_stack = []

    def run_import_module(self, jac_ast):
        """
        import_module:
            KW_IMPORT LBRACE (import_items | '*') RBRACE
            KW_WITH STRING SEMI;

        TODO: Check for duplicate imports and ignore if imported
        """
        kid = jac_ast.kid
        fn = os.path.join(self.builder._mod_dir, parse_str_token(kid[-2].token_text()))
        full_path = os.path.realpath(fn)
        mod_name = os.path.basename(fn)
        mdir = os.path.dirname(full_path) + "/"
        from_mod = self.builder.root.loc[2]
        logger.debug(f"Importing items from {mod_name} to {from_mod}...")
        pre_build = None
        if (mdir + mod_name) in JacAstBuilder._ast_head_map.keys():
            pre_build = JacAstBuilder._ast_head_map[mdir + mod_name]
        elif os.path.isfile(fn):
            with open(fn, "r") as file:
                jac_text = file.read()
            pre_build = JacAstBuilder(
                jac_text=jac_text,
                mod_name=mod_name,
                mod_dir=mdir,
            )
        else:
            err = (
                f"Module not found for import! {mod_name} from" + f" {from_mod} - {fn}"
            )
            self.builder._parse_errors.append(err)
        if pre_build:
            self.builder._parse_errors += pre_build._parse_errors
            self.builder.dependencies += pre_build.dependencies
            import_elements = list(
                filter(lambda x: x.name == "element", pre_build.root.kid)
            )
            if kid[2].name == "STAR_MUL":
                ret = import_elements
            else:
                ret = self.run_import_items(kid[2], import_elements)
            for i in ret:
                self.builder.dependencies.append(i)
            return ret
        else:
            return []

    def run_import_items(self, jac_ast, import_elements):
        """
        import_items:
            WALKER_DBL_COLON (STAR_MUL | import_names) (
                COMMA import_items
            )?
            | NODE_DBL_COLON (STAR_MUL | import_names) (
                COMMA import_items
            )?
            | EDGE_DBL_COLON (STAR_MUL | import_names) (
                COMMA import_items
            )?
            | GRAPH_DBL_COLON (STAR_MUL | import_names) (
                COMMA import_items
            )?
            | KW_GLOBAL DBL_COLON (STAR_MUL | import_names) (
                COMMA import_items
            )?
            | TYPE_DBL_COLON (STAR_MUL | import_names) (
                COMMA import_items
            )?;
        """
        kid = jac_ast.kid
        type_name = (
            "KW_" + kid[0].name.split("_")[0]
            if not kid[0].name.startswith("KW_")
            else kid[0].name
        )
        kid = kid[2:] if kid[1].name == "DBL_COLON" else kid[1:]
        ret_elements = list(
            filter(
                lambda x: x in self.builder.dependencies
                or x.kid[0].kid[0].name == type_name,
                import_elements,
            )
        )
        if kid[0].name == "import_names":
            import_names = self.run_import_names(kid[0])
            ret_elements = list(
                filter(
                    lambda x: x in self.builder.dependencies
                    or x.kid[0].kid[1].token_text() in import_names,
                    ret_elements,
                )
            )
            if len(ret_elements) < len(import_names):
                err = (
                    f"{kid[0].loc[2]}: Line {kid[0].loc[0]}: "
                    + "Module name not found!"
                )
                self.builder._parse_errors.append(err)

        if kid[-1].name == "import_items":
            return ret_elements + self.run_import_items(kid[-1], import_elements)

        return ret_elements

    def run_import_names(self, jac_ast):
        """
        import_names: NAME | LBRACE name_list RBRACE;
        """
        kid = jac_ast.kid
        if kid[0].name == "NAME":
            return [kid[0].token_text()]
        else:
            return self.run_name_list(kid[1])

    def run_name_list(self, jac_ast):
        """
        name_list: NAME (COMMA NAME)*;
        """
        kid = jac_ast.kid
        ret = []
        for i in kid:
            if i.name == "NAME":
                ret.append(i.token_text())
        return ret

    def enterEveryRule(self, ctx):  # noqa
        """Visits every node in antlr parse tree"""
        if len(self.node_stack) == 0:
            new_node = self.builder.root
        else:
            new_node = Ast(mod_name=self.builder.root.loc[2])
        new_node.name = jacParser.ruleNames[ctx.getRuleIndex()]
        new_node.loc[0] = ctx.start.line
        if new_node.name == "architype":
            new_node.src = self.get_code(ctx)
        new_node.loc[1] = ctx.start.column

        if len(self.node_stack) and new_node.name != "import_module":
            self.node_stack[-1].kid.append(new_node)
        self.node_stack.append(new_node)

    def exitEveryRule(self, ctx):  # noqa
        """Overloaded function that visits every node on exit"""
        top = self.node_stack.pop()
        if top.name == "import_module":
            for i in self.run_import_module(top):
                if i not in self.node_stack[-1].kid:
                    self.node_stack[-1].kid.append(i)

    def visitTerminal(self, node):  # noqa
        """Visits terminals as walker walks, adds ast node"""
        new_node = Ast(mod_name=self.builder.root.loc[2])
        new_node.name = jacParser.symbolicNames[node.getSymbol().type]
        new_node.loc[0] = node.getSymbol().line
        new_node.loc[1] = node.getSymbol().column

        token = {
            "symbol": jacParser.symbolicNames[node.getSymbol().type],
            "text": node.getSymbol().text,
        }
        new_node.loc[3]["token"] = token

        self.node_stack[-1].kid.append(new_node)

    def get_code(self, ctx):
        out = ""
        lines = self.code.split("\n")

        for idx, i in enumerate(lines[ctx.start.line - 1 : ctx.stop.line]):
            out += i
            if idx != ctx.stop.line - 1:
                out += "\n"

        return out


class JacTreeError(ErrorListener):
    """Accumulate errors as parse tree is walked"""

    def __init__(self, builder):
        self.builder = builder

    def syntaxError(  # noqa
        self, recognizer, offendingSymbol, line, column, msg, e  # noqa
    ):
        """Add error to error list"""
        root = self.builder.root
        self.builder._parse_errors.append(
            f"{str(root.loc[2])}: line {str(line)}:" f"{str(column)} - {root} - {msg}"
        )



"""Built in actions for Jaseci"""
import random

# import faker
from datetime import datetime
from datetime import timedelta
from jaseci.jsorc.live_actions import jaseci_action

lorem_words = (
    "adipisci aliquam amet consectetur dolor dolore dolorem eius "
    "est et incidunt ipsum labore magnam modi neque non numquam "
    "porro quaerat qui quia quisquam sed sit tempora ut velit "
    "voluptatem"
).split()


@jaseci_action()
def seed(val: int):
    """Seed random num generator"""
    random.seed(val, version=2)
    # faker.Faker.seed(param_list[0])


@jaseci_action()
def integer(start: int, end: int):
    """Random integeter between range"""
    return random.randint(start, end)


@jaseci_action()
def choice(lst: list):
    """Random select and return item in list"""
    return random.choice(lst)


@jaseci_action()
def uniform(low: float, high: float):
    """Random select a float between low an high"""
    return random.uniform(low, high)


@jaseci_action()
def sentence(min_lenth: int = 4, max_length: int = 10, sep: str = " "):
    """Get a random sentence"""
    n = random.randint(min_lenth, max_length)
    s = sep.join(word() for _ in range(n))
    return s[0].upper() + s[1:] + "."


@jaseci_action()
def paragraph(min_lenth: int = 4, max_length: int = 8, sep: str = " "):
    """Get a random paragraph"""
    n = random.randint(min_lenth, max_length)
    p = sep.join(sentence() for _ in range(n))
    return p


@jaseci_action()
def text(min_lenth: int = 3, max_length: int = 6, sep: str = "\n\n"):
    """Get a random text"""
    n = random.randint(min_lenth, max_length)
    t = sep.join(paragraph() for _ in range(n))
    return t


@jaseci_action()
def word():
    """Get a random sentence"""
    return random.choice(lorem_words)


@jaseci_action()
def time(start_date: str, end_date: str):
    """Provide a random datetime between range"""
    start = datetime.fromisoformat(start_date)
    end = datetime.fromisoformat(end_date)
    return (
        start
        + timedelta(
            seconds=random.randint(0, int((end - start).total_seconds())),
        )
    ).isoformat()



import re
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.kube_svc import KubeService
from jaseci.utils.utils import logger, app_logger
from requests import get, post
from datetime import datetime
from copy import copy
from base64 import b64encode
import threading
import queue
import logging.handlers


#################################################
#                  ELASTIC APP                  #
#################################################


@JsOrc.service(name="elastic", config="ELASTIC_CONFIG", manifest="ELASTIC_MANIFEST")
class ElasticService(JsOrc.CommonService):
    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self):
        kube = JsOrc.svc("kube", KubeService)
        if kube.is_running():
            elasticsearches = kube.resolve_manifest(
                self.manifest, *JsOrc.overrided_namespace("elastic", self.manifest_type)
            ).get("Elasticsearch", [])

            if elasticsearches:
                metadata: dict = elasticsearches["jaseci"]["metadata"]

                cert = kube.get_secret(
                    f'{metadata.get("name")}-es-http-certs-internal',
                    "ca.crt",
                    metadata.get("namespace"),
                )

                if cert:
                    with open("elastic-certificate.crt", "w") as cert_file:
                        cert_file.write(cert)

                if not self.config.get("auth"):
                    auth = kube.get_secret(
                        f'{metadata.get("name")}-es-elastic-user',
                        "elastic",
                        metadata.get("namespace"),
                    )
                    self.config[
                        "auth"
                    ] = f'basic {b64encode(f"elastic:{auth}".encode()).decode()}'

        self.app = Elastic(self.config)
        self.app.health("timeout=1s")

    def post_run(self):
        self.add_elastic_log_handler(logger, "core")
        self.add_elastic_log_handler(app_logger, "app")

    def add_elastic_log_handler(self, logger_instance, index):
        has_queue_handler = any(
            isinstance(h, logging.handlers.QueueHandler)
            for h in logger_instance.handlers
        )
        if not has_queue_handler:
            log_queue = queue.Queue()
            queue_handler = logging.handlers.QueueHandler(log_queue)
            logger_instance.addHandler(queue_handler)

            def elastic_log_worker():
                while True:
                    try:
                        record = log_queue.get()
                        if record is None:
                            break
                        # Strip out color code from message before sending to elastic
                        msg = record.getMessage()
                        msg = re.sub(r"\033\[[0-9]*m", "", msg)

                        elastic_record = {
                            "@timestamp": logging.Formatter().formatTime(
                                record, "%Y-%m-%d %H:%M:%S"
                            ),
                            "message": msg,
                            "level": record.levelname,
                        }
                        extra_fields = record.__dict__.get("extra_fields", [])
                        elastic_record.update(
                            dict([(k, record.__dict__[k]) for k in extra_fields])
                        )
                        self.app.doc(log=elastic_record, index=index)
                    except Exception:
                        pass

            worker_thread = threading.Thread(target=elastic_log_worker, daemon=True)
            worker_thread.start()


class Elastic:
    def __init__(self, config: dict):
        if not config.get("url"):
            raise Exception("URL is required!")
        self.url = config["url"]

        if not config.get("common_index"):
            raise Exception("Common Index is required!")
        self.common_index = config["common_index"]

        if not config.get("activity_index"):
            raise Exception("Activity Index is required!")
        self.activity_index = config["activity_index"]

        self.headers = {"Content-Type": "application/json"}

        if config["auth"]:
            self.headers["Authorization"] = config["auth"]

    def _get(self, url: str, json: dict = None):
        return get(
            f"{self.url}{url}",
            json=json,
            headers=self.headers,
            verify="elastic-certificate.crt",
        ).json()

    def _post(self, url: str, json: dict = None):
        return post(
            f"{self.url}{url}",
            json=json,
            headers=self.headers,
            verify="elastic-certificate.crt",
        ).json()

    def post(self, url: str, body: dict, index: str = "", suffix: str = ""):
        return self._post(f"/{index or self.common_index}{suffix}{url}", body)

    def post_act(self, url: str, body: dict, suffix: str = ""):
        return self.post(url, body, self.activity_index, suffix)

    def get(self, url: str, body: dict, index: str = "", suffix: str = ""):
        return self._get(f"/{index or self.common_index}{suffix}{url}", body)

    def get_act(self, url: str, body: dict, suffix: str = ""):
        return self.get(url, body, self.activity_index, suffix)

    def doc(self, log: dict, query: str = "", index: str = "", suffix: str = ""):
        return self.post(f"/_doc?{query}", log, index, suffix)

    def doc_activity(self, log: dict, query: str = "", suffix: str = ""):
        return self.doc(log, query, self.activity_index, suffix)

    def search(self, body: dict, query: str = "", index: str = "", suffix: str = ""):
        return self.get(f"/_search?{query}", body, index, suffix)

    def search_activity(self, body: dict, query: str = "", suffix: str = ""):
        return self.search(body, query, self.activity_index, suffix)

    def mapping(self, query: str = "", index: str = "", suffix: str = ""):
        return self.get(f"/_mapping?{query}", None, index, suffix)

    def mapping_activity(self, query: str = "", suffix: str = ""):
        return self.mapping(query, self.activity_index, suffix)

    def refresh(self, index: str = "", suffix: str = ""):
        return self.get(f"/_refresh", None, index, suffix)

    def refresh_activity(self, suffix: str = ""):
        return self.refresh(self.activity_index, suffix)

    def aliases(self, query: str = "pretty=true"):
        return self._get(f"/_aliases?{query}")

    def reindex(self, body: dict, query: str = "pretty"):
        return self._post(f"/_reindex?{query}", body)

    # standard methods
    def generate_from_meta(self, meta: dict, override: dict, action: str = None):
        scope = meta["scope"].local_scope
        interp = meta["interp"]

        node = interp.current_node
        walker = scope["visitor"]

        override_misc = override.get("misc")
        override["misc"] = {
            "report": copy(interp.report),
            "node": node.serialize(detailed=False),
        }

        if override_misc and isinstance(override_misc, dict):
            override["misc"].update(override_misc)

        master = meta["h"].get_obj(meta["m_id"], meta["m_id"]).master_self(True)

        headers = interp.request_context.get("headers", {})
        if headers.get("Authorization"):
            del headers["Authorization"]

        activity = {
            "datetime": datetime.utcnow().isoformat(),
            "activity_action": action or walker.name.replace("_", " ").title(),
            "activity_type": walker.name,
            "activity_point": node.name,
            "walker_id": walker.jid,
            "node_id": node.jid,
            "master_id": master["jid"],
            "user": master.get("__meta__") or {"email": master["name"]},
            "request_context": interp.request_context,
            "data": walker.context,
        }

        activity.update(override)

        return activity

    def generate_from_request(self, request):
        headers = dict(request.headers)
        if headers.get("Authorization"):
            del headers["Authorization"]

        data = request.data.dict() if type(request.data) is not dict else request.data
        password = data.get("password")
        if password:
            data["password"] = len(password) * "*"

        user = request.user

        return {
            "datetime": datetime.utcnow().isoformat(),
            "activity_action": "User Manage",
            "activity_type": "user_manage",
            "master_id": user.master.urn,
            "user": {"email": user.email},
            "request_context": {
                "method": request.method,
                "headers": headers,
                "query": request.GET.dict(),
                "body": data,
            },
            "data": data,
        }

    def health(self, query: str = ""):
        if self._get(f"/_cluster/health?{query}").get("timed_out", True):
            raise Exception("Cannot connect on elastic service!")



from enum import IntEnum, auto
from jaseci.prim.edge import Edge
from jaseci.prim.node import Node


class JsOp(IntEnum):
    PUSH_SCOPE = auto()
    POP_SCOPE = auto()
    ADD = auto()
    SUBTRACT = auto()
    MULTIPLY = auto()
    DIVIDE = auto()
    MODULO = auto()
    POWER = auto()
    NEGATE = auto()
    COMPARE = auto()
    AND = auto()
    OR = auto()
    ASSIGN = auto()
    COPY_FIELDS = auto()
    REPORT = auto()
    ACTION_CALL = auto()
    INCREMENT = auto()  # [type]
    LOAD_CONST = auto()  # [type, bytes, (val)] / [type, (val)] / [type=type, type]
    LOAD_VAR = auto()  # [bytes, (name)]
    CREATE_VAR = auto()  # [bytes, (name)]
    DEBUG_INFO = auto()  # [bytes, line, bytes, (jacfile)]


class JsType(IntEnum):
    TYPE = auto()
    INT = auto()
    FLOAT = auto()
    STRING = auto()
    LIST = auto()
    DICT = auto()
    BOOL = auto()
    NULL = auto()
    NODE = auto()
    EDGE = auto()


class JsCmp(IntEnum):
    EE = auto()
    LT = auto()
    GT = auto()
    LTE = auto()
    GTE = auto()
    NE = auto()
    IN = auto()
    NIN = auto()
    NOT = auto()
    PEQ = auto()
    MEQ = auto()
    TEQ = auto()
    DEQ = auto()


type_map = {
    JsType.TYPE: type,
    JsType.INT: int,
    JsType.FLOAT: float,
    JsType.STRING: str,
    JsType.LIST: list,
    JsType.DICT: dict,
    JsType.BOOL: bool,
    JsType.NULL: None,
    JsType.NODE: Node,
    JsType.EDGE: Edge,
}

cmp_op_map = {
    JsCmp.NOT: lambda val: not val,
    JsCmp.EE: lambda v1, v2: v1 == v2,
    JsCmp.LT: lambda v1, v2: v1 < v2,
    JsCmp.GT: lambda v1, v2: v1 > v2,
    JsCmp.LTE: lambda v1, v2: v1 <= v2,
    JsCmp.GTE: lambda v1, v2: v1 >= v2,
    JsCmp.NE: lambda v1, v2: v1 != v2,
    JsCmp.IN: lambda v1, v2: v1 in v2,
    JsCmp.NIN: lambda v1, v2: v1 not in v2,
}



"""
ID list class for Jaseci

Generalized functions for managing '_ids' convention for lists of Jaseci
objects

parent_obj is the instance that the list belongs to
"""
from jaseci.utils.utils import logger


class IdList(list):
    """
    ID list class for tracking lists of objects in Jaseci

    ingest_list is a list of hex strings to convert to UUID and append.
    """

    def __init__(self, parent_obj, auto_save=True, in_list=None):
        self.parent_obj = parent_obj
        self.cached_objects = []
        self.heal_list = []
        self.auto_save = auto_save
        if in_list:
            self.extend(in_list)

    def cache_reset(self):
        self.cached_objects = []

    def add_obj(self, obj, push_front=False, allow_dups=False, silent=False):
        """Adds a obj obj to Jaseci object"""
        self.parent_obj.check_hooks_match(obj)
        if not allow_dups and obj.jid in self:
            if not silent:
                logger.warning(str(f"{obj} is already in {self.parent_obj}'s list"))
        else:
            self.cache_reset()
            if push_front:
                self.insert(0, obj.jid)
            else:
                self.append(obj.jid)
            if not obj.j_parent:
                obj.j_parent = self.parent_obj.jid
                self.save(obj)
            self.save()

    def add_obj_list(self, obj_list, push_front=False, allow_dups=False, silent=False):
        self.cache_reset()
        if push_front:
            obj_list.reverse()
        for i in obj_list:
            self.add_obj(i, push_front=push_front, allow_dups=allow_dups, silent=silent)

    def remove_obj(self, obj):
        """Remove a Jaseci obj from list"""
        self.cache_reset()
        self.remove(obj.jid)
        self.save()

    def heal(self):
        for i in self.heal_list:
            self.remove(i)
        if len(self.heal_list) and hasattr(self.parent_obj, "save"):
            self.save()
        self.heal_list = []

    def destroy_obj(self, obj):
        """Completely destroys a Jaseci obj obj by it's name"""
        self.remove_obj(obj)
        obj.destroy()

    def obj_for_id_not_exist_error(self, item_id):
        self.heal_list.append(item_id)
        my_name = "id_list"
        for k, v in self.parent_obj.__dict__.items():
            if id(v) == id(self):
                my_name = k
        return f"{item_id} not found in {my_name} of {self.parent_obj}!"

    def get_obj_by_name(self, name, kind=None, silent=False):
        """Returns a Jaseci obj obj by it's name"""
        ret = None
        for i in self:
            obj = self.parent_obj._h.get_obj(self.parent_obj._m_id, i)
            if not obj:
                logger.critical(self.obj_for_id_not_exist_error(i))
                continue
            if obj.name == name:
                if kind and obj.kind != kind:
                    continue
                ret = obj
                break
        if not ret and not silent:
            logger.error(str(f"object for '{name}' not found in '{self.parent_obj}'!"))
        self.heal()
        return ret

    def has_obj_by_name(self, name, kind=None):
        """Returns whether a Jaseci obj exists by it's name"""
        return self.get_obj_by_name(name, kind, silent=True) is not None

    def remove_obj_by_name(self, name, kind=None):
        """Remove a Jaseci obj by it's name"""
        self.remove_obj(self.get_obj_by_name(name, kind))

    def destroy_obj_by_name(self, name, kind=None):
        """Destroy a Jaseci obj by it's name"""
        self.destroy_obj(self.get_obj_by_name(name, kind))

    def obj_list(self):
        """Return list of objects from ids"""
        if not len(self.cached_objects):
            for i in self:
                obj = self.parent_obj._h.get_obj(self.parent_obj._m_id, i)
                if not obj:
                    logger.critical(self.obj_for_id_not_exist_error(i))
                else:
                    self.cached_objects.append(obj)
        self.heal()
        return self.cached_objects.copy()

    def remove_all(self):
        """Remove a Jaseci obj obj by it's name"""
        for i in self.obj_list():
            self.remove_obj(i)
        if len(self):
            logger.critical(
                str(
                    f"Remove all failed in id_list of {self.parent_obj} - "
                    + f"still has {self}!"
                )
            )

    def destroy_all(self):
        """Remove a Jaseci obj obj by it's name"""
        for i in self.obj_list():
            self.destroy_obj(i)
        if len(self):
            logger.critical(
                str(
                    f"Destroy all failed in id_list of {self.parent_obj} - "
                    + f"still has {self}!"
                )
            )

    def first_obj(self):
        """Get first object in list"""
        if not self:
            logger.error(str(f"List in '{self.parent_obj}' is empty!"))
            return None
        return self.parent_obj._h.get_obj(self.parent_obj._m_id, self[0])

    def pop_first_obj(self):
        """Get first object in list"""
        ret = self.first_obj()
        if ret:
            self.remove_obj(ret)
        return ret

    def save(self, obj=None):
        if self.auto_save:
            self.parent_obj.save()
        if obj:
            obj.save()
